import os
import math
import torch
import logging

from abc import ABC
from sys import platform
from pathlib import Path
from tempfile import TemporaryDirectory
from torch.utils.data import DataLoader
from contextlib import nullcontext as _nullcontext
from safetensors.torch import load_file, save_file
from typing import Callable, Optional, Union, Dict, List, Any

from .metrics_calculator import MetricsCalculator
from .solver_utils import get_mean_and_confidence_bounds

from ..models import BiotrainerModel
from ..output_files import OutputManager
from ..utilities import get_logger, EpochMetrics

logger = get_logger(__name__)


class Solver(ABC):

    def __init__(self,
                 # Necessary
                 split_name, protocol, network: BiotrainerModel, optimizer,
                 loss_function, metrics_calculator: MetricsCalculator,
                 # Optional with defaults
                 output_manager: Optional[OutputManager] = None,
                 log_dir: str = "",
                 number_of_epochs: int = 1000, patience: int = 20, epsilon: float = 0.001,
                 device: Union[None, str, torch.device] = None,
                 # Used by classification subclasses
                 n_classes: Optional[int] = 0):

        self.split_name = split_name
        self.checkpoint_type = "safetensors"
        self.checkpoint_name = f"{self.split_name}_checkpoint.{self.checkpoint_type}"
        self.protocol = protocol
        self.network = network
        self.optimizer = optimizer
        self.loss_function = loss_function
        self.metrics_calculator = metrics_calculator

        self.output_manager = output_manager
        self.start_epoch = 0
        self.number_of_epochs = number_of_epochs
        self.patience = patience
        self.epsilon = epsilon
        self.log_dir = log_dir

        # Early stopping internal variables
        self._best_epoch = 0
        self._min_loss = math.inf
        self._stop_count = patience
        self._tempdir = TemporaryDirectory()

        # Device handling
        self.device = device
        self.network = network.to(self.device)

    def __del__(self):
        self._tempdir.cleanup()

    def train(self, training_dataloader: DataLoader, validation_dataloader: DataLoader) -> List[EpochMetrics]:
        # Get things ready
        self.network = self.network.train()
        self._min_loss = math.inf
        epoch_iterations: List[EpochMetrics] = []

        # Make an initial save of the model if trained from scratch
        if self.start_epoch == 0:
            self.save_checkpoint(0)

        for epoch in range(self.start_epoch, self.number_of_epochs):

            # Evaluating before testing: This way val_loss > train_loss holds for most epochs
            # If we would train before validating, the validation would benefit from the knowledge gained during
            # training, thus most likely val_loss < train_loss would be true for most epochs (and a bit confusing)
            validation_iterations = list()
            for i, (_, X, y, lengths) in enumerate(validation_dataloader):
                iteration_result = self._training_iteration(
                    X, y, step=len(epoch_iterations) * len(validation_dataloader) + len(validation_iterations) + 1,
                    context=torch.no_grad, lengths=lengths
                )
                validation_iterations.append(iteration_result)
            validation_epoch_metrics = self.metrics_calculator.compute_metrics()

            train_iterations = list()
            for i, (_, X, y, lengths) in enumerate(training_dataloader):
                iteration_result = self._training_iteration(
                    X, y, step=len(epoch_iterations) * len(training_dataloader) + len(train_iterations) + 1,
                    lengths=lengths
                )
                train_iterations.append(iteration_result)
            train_epoch_metrics = self.metrics_calculator.compute_metrics()

            epoch_metrics = EpochMetrics(epoch=epoch,
                                         training={**Solver._aggregate_iteration_losses(train_iterations),
                                                   **train_epoch_metrics},
                                         validation={**Solver._aggregate_iteration_losses(validation_iterations),
                                                     **validation_epoch_metrics}, )

            epoch_iterations.append(epoch_metrics)

            # Logging
            logger.info(f"Epoch {epoch}")
            logger.info(f"Training results")
            for key in epoch_metrics.training:
                logger.info(f"\t{key}: {epoch_metrics.training[key]:.2f}")
            logger.info(f"Validation results")
            for key in epoch_metrics.validation:
                logger.info(f"\t{key}: {epoch_metrics.validation[key]:.2f}")

            if self.output_manager:
                self.output_manager.add_training_iteration(split_name=self.split_name, epoch_metrics=epoch_metrics)

            if self._early_stop(current_loss=epoch_metrics.validation['loss'], epoch=epoch):
                logger.info(f"Early stopping triggered - patience limit is reached! (Best epoch: {self._best_epoch})")
                return epoch_iterations

        return epoch_iterations

    def inference(self, dataloader: DataLoader, calculate_test_metrics: bool = False) -> \
            Dict[str, Union[List[Any], Dict[str, Union[float, int]]]]:
        self.network = self.network.eval()

        predict_iterations = list()
        mapped_predictions = dict()
        mapped_probabilities = dict()

        for i, (seq_ids, X, y, lengths) in enumerate(dataloader):
            if calculate_test_metrics:  # For test set, y must be valid targets
                iteration_result = self._training_iteration(
                    X, y, context=torch.no_grad, lengths=lengths
                )
            else:  # For new predictions, y is ignored
                iteration_result = self._prediction_iteration(x=X, lengths=lengths)

            predict_iterations.append(iteration_result)
            # Create dict with seq_id: prediction
            for idx, prediction in enumerate(iteration_result["prediction"]):
                mapped_predictions[seq_ids[idx]] = prediction

            for idx, probability in enumerate(iteration_result["probabilities"]):
                mapped_probabilities[seq_ids[idx]] = probability

        metrics = None
        if calculate_test_metrics:
            metrics = {**Solver._aggregate_iteration_losses(predict_iterations),
                       **self.metrics_calculator.compute_metrics()
                       }

        return {
            'metrics': metrics,
            'mapped_predictions': mapped_predictions,
            'mapped_probabilities': mapped_probabilities
        }

    @staticmethod
    def _enable_dropout(model):
        """ Function to enable the dropout layers during test-time """
        number_dropout_layers = 0
        for m in model.modules():
            if m.__class__.__name__.startswith('Dropout'):
                m.train()
                if m.p > 0.0:
                    number_dropout_layers += 1
        if not number_dropout_layers > 0:
            raise ValueError("Trying to do monte carlo dropout inference on model without dropout!")

    def _do_dropout_iterations(self, X, lengths, n_forward_passes):
        dropout_iterations = []
        for idx_forward_pass in range(n_forward_passes):
            self.network = self.network.eval()
            self._enable_dropout(self.network)
            dropout_iteration_result = self._prediction_iteration(x=X, lengths=lengths)
            dropout_iterations.append(dropout_iteration_result)

        return dropout_iterations

    def inference_monte_carlo_dropout(self, dataloader: DataLoader,
                                      n_forward_passes: int = 30,
                                      confidence_level: float = 0.05):
        """
        Calculate inference results from existing models for given embeddings

            dataloader: Dataloader with embeddings
            n_forward_passes: Times to repeat calculation with different dropout nodes enabled
            confidence_level: Confidence level for result confidence intervals (0.05 => 95% percentile)
        """
        if not 0 < confidence_level < 1:
            raise ValueError(f"Confidence level must be between 0 and 1, given: {confidence_level}!")

        mapped_predictions = {}

        for i, (seq_ids, X, y, lengths) in enumerate(dataloader):
            dropout_iterations = self._do_dropout_iterations(X, lengths, n_forward_passes)

            dropout_raw_values = torch.stack([torch.tensor(dropout_iteration["probabilities"])
                                              for dropout_iteration in dropout_iterations], dim=1)

            dropout_mean, lower_bound, upper_bound = get_mean_and_confidence_bounds(values=dropout_raw_values,
                                                                            dimension=1,
                                                                            confidence_level=confidence_level)
            prediction_by_mean = self._probabilities_to_predictions(dropout_mean)

            # Create dict with seq_id: prediction
            for idx, prediction in enumerate(prediction_by_mean):
                mapped_predictions[seq_ids[idx]] = {"prediction": prediction.item(),
                                                    "all_predictions": [dropout_iteration["prediction"][idx] for
                                                                        dropout_iteration in dropout_iterations],
                                                    "mcd_mean": dropout_mean[idx],
                                                    "mcd_lower_bound": (
                                                            lower_bound[idx]),
                                                    "mcd_upper_bound": (
                                                            upper_bound[idx]),
                                                    }

        return {
            'mapped_predictions': mapped_predictions
        }

    def auto_resume(self, training_dataloader: DataLoader, validation_dataloader: DataLoader,
                    train_wrapper):
        available_checkpoints = [checkpoint for checkpoint in os.listdir(self.log_dir)
                                 if checkpoint.split("_")[0] == self.checkpoint_name.split("_")[0]]
        if self.checkpoint_name in available_checkpoints:
            # Hold out
            if "hold_out" in self.checkpoint_name:
                self.load_checkpoint(resume_training=True)
                return train_wrapper(self)
            # K_fold and leave_p_out: Checkpoint is already there, now check if it has been trained completely
            # Sort by name and by last split number (differs from alphabetical order because str(9) > str(10))
            available_checkpoints_sorted = sorted(available_checkpoints,
                                                  key=lambda ch_pt: (ch_pt.split("_checkpoint.")[0].split("-")[0:-1],
                                                                     int(ch_pt.split("_checkpoint.")[0].split("-")[-1]
                                                                         )
                                                                     )
                                                  )
            if available_checkpoints_sorted.index(self.checkpoint_name) < (len(available_checkpoints_sorted) - 1):
                # Do inference on train and val to get metrics of training
                self.load_checkpoint(resume_training=False)
                train_set_result = self.inference(training_dataloader, calculate_test_metrics=True)
                val_set_result = self.inference(validation_dataloader, calculate_test_metrics=True)
                return {
                    "training": train_set_result["metrics"],
                    "validation": val_set_result["metrics"],
                    "epoch": self.start_epoch
                }

        # Checkpoint not available or not sure if checkpoint training has finished
        logger.info(f"No pretrained checkpoint ({self.checkpoint_name}) found for auto_resume. "
                    f"Training new model from scratch!")
        return train_wrapper(self)

    def load_checkpoint(self, checkpoint_path: Path = None, resume_training: bool = False, disable_pytorch_compile: bool = True):
        if checkpoint_path:
            checkpoint_file = checkpoint_path
            self.checkpoint_type = checkpoint_path.suffix
        elif self.log_dir:
            checkpoint_file = Path(self.log_dir) / self.checkpoint_name
        else:
            checkpoint_file = Path(self._tempdir.name) / self.checkpoint_name

        if checkpoint_file.suffix == '.safetensors':
            # Load safetensors checkpoint
            state_dict = load_file(str(checkpoint_file))
            state = {
                'state_dict': state_dict,
                'epoch': state_dict.pop('epoch').item()
            }
        else:
            # Load PyTorch checkpoint is deprecated since v1.0.0
            raise Exception("Cannot load pt checkpoint because torch_pt_loading is not allowed since v1.0.0!")
        try:
            downstream_model = self.network.get_downstream_model()
            if not disable_pytorch_compile:
                downstream_model.compile()
            downstream_model.load_state_dict(state['state_dict'])

            if resume_training:
                self.start_epoch = state['epoch'] + 1
                if self.start_epoch == self.number_of_epochs:
                    raise Exception(f"Cannot resume training of checkpoint because model has already "
                                    f"been trained for maximum number_of_epochs ({self.number_of_epochs})!")
            else:
                self.start_epoch = state['epoch']
            self.network = self.network.to(self.device)
            logger.info(f"Loaded model from epoch: {state['epoch']}")
        except RuntimeError as e:
            raise Exception(f"Defined model architecture does not seem to match pretrained model!") from e

    def get_best_epoch(self) -> int:
        if self.start_epoch > 0:
            return self._best_epoch - self.start_epoch
        else:
            return self._best_epoch

    def save_checkpoint(self, epoch: int):
        # Prepare the state dictionary
        state = {
            'epoch': torch.tensor(epoch),
            **self.network.get_downstream_model().state_dict(),
        }

        # Flatten optimizer state
        optimizer_state = self.optimizer.state_dict()
        for k, v in optimizer_state.items():
            if isinstance(v, dict):
                for inner_k, inner_v in v.items():
                    if isinstance(inner_v, torch.Tensor):
                        state[f'optimizer.{k}.{inner_k}'] = inner_v.contiguous()
            elif isinstance(v, torch.Tensor):
                state[f'optimizer.{k}'] = v.contiguous()

        # Ensure all tensors are contiguous
        state = {k: v.contiguous() if isinstance(v, torch.Tensor) else v for k, v in state.items()}

        if self.log_dir:
            save_path = Path(self.log_dir) / self.checkpoint_name
        else:
            save_path = Path(self._tempdir.name) / self.checkpoint_name

        save_file(state, str(save_path))
        # Log checkpoint path on start of training
        if epoch == 0:
            logger.info(f"Checkpoint(s) will be stored at {save_path}")

    def save_as_onnx(self, embedding_dimension: int, output_dir: Optional[str] = None) -> str:
        output_dir = output_dir if output_dir is not None else self.log_dir

        dummy_input = self.protocol.get_dummy_input(embedding_dimension).to(self.device)

        # Use eval mode during export to avoid batch size problems
        self.network.eval()

        # Export
        onnx_file_name = f"{output_dir}/{self.checkpoint_name.split('.')[0]}.onnx"

        self._onnx_export_strategy(network=self.network.get_downstream_model(),
                                   dummy_input=dummy_input,
                                   onnx_file_name=onnx_file_name)

        return onnx_file_name

    @staticmethod
    def _onnx_export_strategy(network, dummy_input, onnx_file_name: str):
        if "win" in platform.lower():
            torch.onnx.export(
                network,
                dummy_input,
                onnx_file_name,
                input_names=['input'],
                output_names=['output'],
                dynamic_axes={
                    'input': {
                        0: 'batch_size',  # dynamic batch size
                        1: 'sequence_length'  # dynamic sequence length
                    },
                    'output': {
                        0: 'batch_size',
                        1: 'sequence_length'
                    }
                },
            )
        else:
            export_options = torch.onnx.ExportOptions(dynamic_shapes=True,
                                                      diagnostic_options=torch.onnx.DiagnosticOptions(
                                                          verbosity_level=logging.ERROR))
            onnx_program = torch.onnx.dynamo_export(network, dummy_input,
                                                    export_options=export_options)
            onnx_program.save(onnx_file_name)

    def _early_stop(self, current_loss: float, epoch: int) -> bool:
        if current_loss < (self._min_loss - self.epsilon):
            self._min_loss = current_loss
            self._stop_count = self.patience
            self._best_epoch = epoch
            # Save best model (overwrite if necessary)
            self.save_checkpoint(epoch)
            return False
        else:
            if self._stop_count == 0:
                # Trigger early stop
                return True
            else:
                self._stop_count = self._stop_count - 1
                return False

    @staticmethod
    def _aggregate_iteration_losses(iteration_results) -> Dict[str, Any]:
        mean_loss = sum([i["loss"] for i in iteration_results]) / len(iteration_results)
        return {"loss": mean_loss}

    def _transform_network_output(self, network_output: torch.Tensor) -> torch.Tensor:
        """
        Transform network_output shape if necessary.

        :param network_output: The network_output from the ML model employed
        :return: A torch tensor of the transformed network_output
        """

        return network_output

    def _logits_to_probabilities(self, logits: torch.Tensor) -> torch.Tensor:
        """
        An optional transform function which goes from logits to probabilities (softmax)
        Regression tasks simply ignore this step

        :param logits: The logits from the ML model employed
        :return: A torch tensor of the transformed logits (torch.softmax)
        """
        return logits

    def _probabilities_to_predictions(self, probabilities: torch.Tensor) -> torch.Tensor:
        """
        An optional transform function which goes from logits or probabilities to predictions (e.g. classes)
        Regression tasks simply ignore this step

        :param probabilities: The probabilities from the transformed logits of the ML model employed
        :return: A torch tensor of the discretised probabilities
        """
        return probabilities

    def _training_iteration(
            self, x: torch.Tensor, y: torch.Tensor, step=1, context: Optional[Callable] = None,
            lengths: Optional[torch.LongTensor] = None
    ) -> Dict[str, Union[float, list, Dict[str, Union[float, int]]]]:
        do_loss_propagation = False

        if not context:
            context = _nullcontext
            do_loss_propagation = True

        # Move everything on device
        x = x.to(self.device) if isinstance(x, torch.Tensor) else x
        y = y.to(self.device) if isinstance(y, torch.Tensor) else y

        with context():
            if do_loss_propagation:
                self.optimizer.zero_grad()

            # Targets are necessary for the fine-tuning model
            network_output = self.network(x, targets=y)
            if isinstance(network_output, tuple) and len(network_output) == 2:
                logits = network_output[0]
                y = network_output[1]
            else:
                logits = network_output

            # Apply logit transformations before computing loss
            logits = self._transform_network_output(logits)
            loss = self.loss_function(logits, y)

            # Transform logits to probabilities if necessary
            probabilities = self._logits_to_probabilities(logits)
            # Discretize predictions if necessary
            prediction = self._probabilities_to_predictions(probabilities)
            metrics = self.metrics_calculator.compute_metrics(predicted=prediction, labels=y)

            if do_loss_propagation:
                # Do a forward pass & update weights
                self.optimizer.zero_grad()  # clear gradients for next train
                loss.backward()  # backpropagation, compute gradients
                self.optimizer.step()  # apply gradients

            return {
                'loss': loss.item(),
                'prediction': prediction.tolist(),
                'probabilities': probabilities
            }

    def _prediction_iteration(self, x: torch.Tensor, lengths: Optional[torch.LongTensor] = None) -> \
            Dict[str, Union[List, torch.Tensor]]:

        with torch.no_grad():
            # Move everything on device
            x = x.to(self.device) if isinstance(x, torch.Tensor) else x
            network_output = self.network(x)
            if isinstance(network_output, tuple) and len(network_output) == 2:
                logits = network_output[0]
            else:
                logits = network_output

            # Apply transformations
            logits = self._transform_network_output(logits)
            # Transform logits to probabilities if necessary
            probabilities = self._logits_to_probabilities(logits)
            # Discretize predictions if necessary
            prediction = self._probabilities_to_predictions(probabilities)
            return {"prediction": prediction.tolist(),
                    "probabilities": probabilities.tolist()}



"""
Abstract interface for Embedder.
File adapted from bio_embeddings (https://github.com/sacdallago/bio_embeddings/blob/efb9801f0de9b9d51d19b741088763a7d2d0c3a2/bio_embeddings/embed/embedder_interfaces.py)
Original Authors:
  Christian Dallago (sacdallago)
  Konstantin Schuetze (konstin)
"""

import abc
import torch

from numpy import ndarray
from typing import List, Generator, Optional, Iterable, Any, Union, Callable

from .preprocessing_strategies import preprocess_sequences_without_whitespaces

from ...utilities import get_logger, get_device_memory

logger = get_logger(__name__)


class EmbedderInterface(abc.ABC):
    name: str
    _model: Optional[Any] = None
    _device: Union[None, str, torch.device] = None
    _preprocessing_strategy: Callable = lambda self, sequences, mask_token: preprocess_sequences_without_whitespaces(sequences, mask_token)

    @abc.abstractmethod
    def _embed_single(self, sequence: str) -> ndarray:
        """
        Returns embedding for one sequence.

        :param sequence: Valid amino acid sequence as String
        :return: An embedding of the sequence.
        """

        raise NotImplementedError


    def get_mask_token(self) -> Optional[str]:
        return None

    def _preprocess_sequences(self, sequences: Iterable[str]) -> List[str]:
        return self._preprocessing_strategy(sequences, self.get_mask_token())

    def _embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        """Computes the embeddings from all sequences in the batch

        The provided implementation is dummy implementation that should be
        overwritten with the appropriate batching method for the model."""
        for sequence in batch:
            yield self._embed_single(sequence)

    def estimate_batch_size(self, preprocessed_sequences: List[str]) -> int:
        memory_gb = get_device_memory(self._device)
        safety_factor = 0.8
        residues_per_gb = 1024  # Estimating 1MB/residue
        batch_size = int(memory_gb * residues_per_gb * safety_factor)

        # Account for separators that are processed by the tokenizer (do not count towards batch_size => double it)
        first_sequence = preprocessed_sequences[0]
        if any([separator in first_sequence for separator in (' ', ',', ';')]):
            batch_size *= 2

        return batch_size

    def embed_many(
            self, sequences: Iterable[str], batch_size: Optional[int] = None
    ) -> Generator[ndarray, None, None]:
        """
        Yields embedding for one sequence at a time.

        :param sequences: List of proteins as AA strings
        :param batch_size: For embedders that profit from batching, this is maximum number of AA per batch
        :return: A list object with embeddings of the sequences.
        """
        sequences = self._preprocess_sequences(sequences)

        if batch_size is None:
            batch_size = self.estimate_batch_size(preprocessed_sequences=sequences)

        if batch_size and batch_size > 1:
            batch = []
            length = 0
            for sequence in sequences:
                if len(sequence) > batch_size:
                    logger.warning(
                        f"A sequence is {len(sequence)} residues long, "
                        f"which is longer than your `batch_size` parameter which is {batch_size}"
                    )
                    yield from self._embed_batch([sequence])
                    continue
                if length + len(sequence) >= batch_size:
                    yield from self._embed_batch(batch)
                    batch = []
                    length = 0
                batch.append(sequence)
                length += len(sequence)
            yield from self._embed_batch(batch)
        else:
            for seq in sequences:
                yield self._embed_single(seq)

    @staticmethod
    def reduce_per_protein(embedding: torch.tensor) -> torch.tensor:
        """
        For a variable size embedding, returns a fixed size embedding encoding all information of a sequence.

        :param embedding: the embedding
        :return: A fixed size embedding (a vector of size N, where N is fixed)
        """
        return embedding.mean(axis=0)


class EmbedderWithFallback(EmbedderInterface, abc.ABC):
    """ Batching embedder that will fallback to the CPU if the embedding on the GPU failed """

    _model: Any

    @abc.abstractmethod
    def _embed_batch_implementation(
            self, batch: List[str], model: Any
    ) -> Generator[torch.tensor, None, None]:
        ...

    @abc.abstractmethod
    def _get_fallback_model(self):
        """Returns a (cached) cpu model.

        Note that the fallback models generally don't support half precision mode and therefore ignore
        the `half_precision_model` option (https://github.com/huggingface/transformers/issues/11546).
        """
        ...

    def _embed_batch(self, batch: List[str]) -> Generator[torch.tensor, None, None]:
        """Tries to get the embeddings in this order:
          * Full batch GPU
          * Single Sequence GPU
          * Single Sequence CPU

        Single sequence processing is done in case of runtime error due to
        a) very long sequence or b) too large batch size
        If this fails, you might want to consider lowering batch_size and/or
        cutting very long sequences into smaller chunks

        Returns unprocessed embeddings
        """
        # No point in having a fallback model when the normal model is CPU already
        if self._device.type == "cpu":
            yield from self._embed_batch_implementation(batch, self._model)
            return

        try:
            yield from self._embed_batch_implementation(batch, self._model.to(self._device))
        except RuntimeError as e:
            if len(batch) == 1:
                logger.warning(
                    f"RuntimeError for sequence with {len(batch[0])} residues - falling back to CPU.. "
                    f"The error most likely means that you don't have enough "
                    f"GPU RAM to embed a protein this long on the GPU. Full error: {e}"
                )
                yield from self._embed_batch_implementation(batch, self._get_fallback_model())
            else:
                logger.warning(
                    f"RuntimeError processing batch of {len(batch)} sequences. "
                    f"You might want to consider adjusting the `batch_size` parameter. "
                    f"Will try to embed each sequence in the batch individually on the GPU. Full error: {e}"
                )
                for sequence in batch:
                    try:
                        yield from self._embed_batch_implementation([sequence], self._model.to(self._device))
                    except RuntimeError as e:
                        logger.warning(
                            f"RuntimeError for sequence with {len(sequence)} residues - falling back to CPU.. "
                            f"The error most likely means that you don't have enough "
                            f"GPU RAM to embed a protein this long on the GPU. Full error: {e}"
                        )
                        yield from self._embed_batch_implementation([sequence], self._get_fallback_model())



import os
import tempfile

from ruamel import yaml
from pathlib import Path
from typing import Any, Dict

from biotrainer.utilities.cli import train
from biotrainer.config import ConfigurationException

protocol_to_input = {
    'residue_to_class': {'input_file': "test_input_files/r2c/input.fasta",
                         'loss_choice': "cross_entropy_loss"},
    'residue_to_class-error1': {'input_file': "test_input_files/r2c_error1/sequences.fasta",
                                # Missing target in input file
                                'loss_choice': "cross_entropy_loss"},
    'residue_to_class-error2': {'input_file': "test_input_files/r2c_error2/sequences.fasta",
                                # Sequence and labels length mismatch
                                'loss_choice': "cross_entropy_loss"},
    'residue_to_value': {'input_file': "test_input_files/r2v/input.fasta",
                         'loss_choice': "mean_squared_error"},
    'residues_to_class': {'input_file': "test_input_files/s2c/sequences.fasta",
                          'loss_choice': "cross_entropy_loss"},
    'residues_to_value': {'input_file': "test_input_files/s2v/sequences.fasta",
                          'loss_choice': "mean_squared_error"},
    'sequence_to_class': {'input_file': "test_input_files/s2c/sequences.fasta",
                          'loss_choice': "cross_entropy_loss"},
    'sequence_to_class-interactionmultiply': {'input_file': "test_input_files/ppi/interactions.fasta",
                                              'loss_choice': "cross_entropy_loss",
                                              'interaction': "multiply"},
    'sequence_to_class-interactionconcat': {'input_file': "test_input_files/ppi/interactions.fasta",
                                            'loss_choice': "cross_entropy_loss",
                                            'interaction': "concat"},
    'sequence_to_value': {'input_file': "test_input_files/s2v/sequences.fasta",
                          'loss_choice': "mean_squared_error"},

}


def setup_config(protocol: str, model_choice: str, embedder_name: str, tmp_config_dir: str) -> Dict[str, Any]:
    template_config_file_path = "test_config.yml"
    with open(template_config_file_path, "r") as config_file:
        config: dict = yaml.load(config_file, Loader=yaml.Loader)
        input_path_absolute = str(Path(protocol_to_input[protocol]["input_file"]).absolute())
        config["input_file"] = input_path_absolute
        config["loss_choice"] = protocol_to_input[protocol]["loss_choice"]
        actual_protocol = protocol.split("-")[0]
        if len(protocol.split("-")) > 1 and "interaction" in protocol.split("-")[1]:
            config["interaction"] = protocol_to_input[protocol]["interaction"]
        config["protocol"] = actual_protocol
        config["model_choice"] = model_choice
        config["embedder_name"] = embedder_name
        config["output_dir"] = tmp_config_dir
        config["device"] = "cpu"  # Only use cpu for testing
        return config


def test_protocol_config(protocol: str, model: str, embedder_name: str, should_fail: bool):
    print("TESTING CONFIG: " + protocol + " - " + model +
          " - " + embedder_name + " - Passed when failed: " + str(should_fail))
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        config = setup_config(protocol=protocol,
                              model_choice=model,
                              embedder_name=embedder_name,
                              tmp_config_dir=tmp_dir_name)
        try:
            result = train(config=config)
            assert "test_results" in result, "Result does not contain test set metrics!"
            assert os.path.exists(f"{tmp_dir_name}/out.yml"), "No output file generated, run failed!"
        except ConfigurationException:
            assert should_fail, "A ConfigurationException was thrown although it shouldn't have."
        except Exception:
            assert should_fail, "An exception was thrown although it shouldn't have."



import torch

from .solver import Solver


class SequenceClassificationSolver(Solver):

    def _logits_to_probabilities(self, logits: torch.Tensor) -> torch.Tensor:
        return torch.softmax(logits, dim=1)

    def _probabilities_to_predictions(self, probabilities: torch.Tensor) -> torch.Tensor:
        _, predicted_classes = torch.max(probabilities, dim=1)

        return predicted_classes


class SequenceRegressionSolver(Solver):
    def _transform_network_output(self, network_output: torch.Tensor) -> torch.Tensor:
        return network_output.flatten().float()



import unittest

from biotrainer.trainers.cv_splitter import CrossValidationSplitter


class CrossValidationSplitterTests(unittest.TestCase):
    def test_create_bins_from_continuous_values(self):
        bins1 = CrossValidationSplitter._continuous_values_to_bins([1])
        self.assertTrue(len(bins1) == 1, "Bins creation for single value failed!")

        bins2 = CrossValidationSplitter._continuous_values_to_bins([x for x in range(0, 10000, 1)])
        bins3 = CrossValidationSplitter._continuous_values_to_bins([x for x in range(0, 20000, 2)])
        self.assertTrue(len(set(bins2)) == len(set(bins3)), "Bins set size is not equal, "
                                                            "but it must be equal for all inputs!")

        bins4 = CrossValidationSplitter._continuous_values_to_bins([x for x in range(-10000, 50000, 3)])
        self.assertTrue(len(bins4) == len(list(range(-10000, 50000, 3))), "Bins creation with negative numbers failed!")


if __name__ == '__main__':
    unittest.main()



from typing import Dict, Any

from ..pipeline import PipelineContext, PipelineStep
from ..pipeline.pipeline_step import PipelineStepType

from ...protocols import Protocol
from ...utilities import get_logger
from ...embedders import EmbeddingService

logger = get_logger(__name__)


class ProjectionStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.PROJECTION

    @staticmethod
    def _is_dimension_reduction_possible(context: PipelineContext, dimension_reduction_method, n_reduced_components,
                                         id2emb: Dict[str, Any]) -> bool:
        protocol: Protocol = context.config["protocol"]

        min_number_embeddings = 3
        min_number_dimensions = 3

        number_embeddings = len(id2emb)
        number_dimensions = next(iter(id2emb.values())).shape[0]
        if (protocol.using_per_sequence_embeddings() and dimension_reduction_method and n_reduced_components and
                number_embeddings >= min_number_embeddings and
                number_dimensions >= min_number_dimensions):
            return True
        else:
            if dimension_reduction_method and n_reduced_components:
                # Check for errors
                if number_embeddings < min_number_embeddings:
                    raise ValueError(f"Dimensionality reduction cannot be performed as \
                                the number of samples is less than {min_number_embeddings}")
                if number_dimensions < 3:
                    raise ValueError(f"Dimensionality reduction cannot be performed as \
                                the original embedding dimension is less than {min_number_dimensions}")
                if not protocol.using_per_sequence_embeddings():
                    raise ValueError(f"Dimensionality reduction cannot be performed as \
                                the embeddings are not per-sequence embeddings")
            return False

    def process(self, context: PipelineContext) -> PipelineContext:
        id2emb = context.id2emb
        assert id2emb is not None and len(id2emb) > 0, f"id2emb cannot be None or empty at the projection step!"

        dimension_reduction_method = context.config.get("dimension_reduction_method")
        n_reduced_components = context.config.get("n_reduced_components")

        if self._is_dimension_reduction_possible(context, dimension_reduction_method, n_reduced_components, id2emb):
            id2emb = EmbeddingService.embeddings_dimensionality_reduction(
                embeddings=id2emb,
                dimension_reduction_method=dimension_reduction_method,
                n_reduced_components=n_reduced_components)
        else:
            logger.info(f"No dimension reduction performed (as configured).")

        context.id2emb = id2emb
        return context



from __future__ import annotations

import torch

from abc import ABC, abstractmethod
from typing import Optional, Dict, Union
from torchmetrics import Accuracy, Precision, Recall, F1Score, SpearmanCorrCoef, MatthewsCorrCoef, Metric, \
    MeanSquaredError

from ..utilities import MASK_AND_LABELS_PAD_VALUE


class MetricsCalculator(ABC):
    def __init__(self, device, n_classes: int):
        self.device = device
        self.n_classes = n_classes

    def reset(self) -> MetricsCalculator:
        # Reset all metric attributes that are instances of torchmetrics.Metric
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if isinstance(attr, Metric):
                attr.reset()
        return self

    @abstractmethod
    def compute_metrics(
            self, predicted: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None) -> Dict[str, Union[int, float]]:
        raise NotImplementedError

    @staticmethod
    def _compute_metric(metric: Metric, predicted: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """
        Utility function to calculate metrics either on a per-epoch or a per-batch basis

        :param metric: torchmetrics object
        :param predicted: The predicted label/value for each sample
        :param labels: The actual label for each sample

        :return: metric result calculated via metric object
        """
        if predicted is None and labels is None:
            # Per epoch
            metric_result = metric.compute()
            metric.reset()
            return metric_result
        else:
            # Per batch
            if metric.__class__ == SpearmanCorrCoef:
                # SCC only accepts float tensors
                return metric(predicted.cpu().float(), labels.cpu().float())
            return metric(predicted.cpu(), labels.cpu())


class ClassificationMetricsCalculator(MetricsCalculator):
    def __init__(self, device, n_classes: int):
        super().__init__(device, n_classes)

        task = "multiclass" if self.n_classes > 2 else "binary"

        self.acc = Accuracy(task=task, average="micro", num_classes=self.n_classes)

        self.macro_precision = Precision(task=task, average="macro", num_classes=self.n_classes)
        self.micro_precision = Precision(task=task, average="micro", num_classes=self.n_classes)
        self.precision_per_class = Precision(task=task, average="none", num_classes=self.n_classes)

        self.macro_recall = Recall(task=task, average="macro", num_classes=self.n_classes)
        self.micro_recall = Recall(task=task, average="micro", num_classes=self.n_classes)
        self.recall_per_class = Recall(task=task, average="none", num_classes=self.n_classes)

        self.macro_f1_score = F1Score(task=task, average="macro", num_classes=self.n_classes)
        self.micro_f1_score = F1Score(task=task, average="micro", num_classes=self.n_classes)
        self.f1_per_class = F1Score(task=task, average="none", num_classes=self.n_classes)

        self.scc = SpearmanCorrCoef()
        self.mcc = MatthewsCorrCoef(task=task, num_classes=self.n_classes)

    def compute_metrics(
            self, predicted: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None) -> Dict[str, Union[int, float]]:
        def _compute_metric(metric) -> torch.Tensor:
            # To shorten the code below, this delegate function is used
            return self._compute_metric(metric, predicted=predicted, labels=labels)

        metrics_dict = {'accuracy': _compute_metric(self.acc).item()}

        # Multi-class prediction
        if self.n_classes > 2:
            precision_per_class = _compute_metric(self.precision_per_class)
            precisions = {'- precision class {}'.format(i): precision_per_class[i].item() for i in
                          range(self.n_classes)}
            metrics_dict['macro-precision'] = _compute_metric(self.macro_precision).item()
            metrics_dict['micro-precision'] = _compute_metric(self.micro_precision).item()
            metrics_dict.update(precisions)

            recall_per_class = _compute_metric(self.recall_per_class)
            recalls = {'- recall class {}'.format(i): recall_per_class[i].item() for i in range(self.n_classes)}
            metrics_dict['macro-recall'] = _compute_metric(self.macro_recall).item()
            metrics_dict['micro-recall'] = _compute_metric(self.micro_recall).item()
            metrics_dict.update(recalls)

            f1_per_class = _compute_metric(self.f1_per_class)
            f1scores = {'- f1_score class {}'.format(i): f1_per_class[i].item() for i in range(self.n_classes)}
            metrics_dict['macro-f1_score'] = _compute_metric(self.macro_f1_score).item()
            metrics_dict['micro-f1_score'] = _compute_metric(self.micro_f1_score).item()
            metrics_dict.update(f1scores)
        # Binary prediction
        else:
            metrics_dict['precision'] = _compute_metric(self.macro_precision).item()
            metrics_dict['recall'] = _compute_metric(self.macro_recall).item()
            metrics_dict['f1_score'] = _compute_metric(self.macro_f1_score).item()

        metrics_dict['spearmans-corr-coeff'] = _compute_metric(self.scc).item()
        metrics_dict['matthews-corr-coeff'] = _compute_metric(self.mcc).item()

        return metrics_dict


class RegressionMetricsCalculator(MetricsCalculator):
    def __init__(self, device, n_classes: int):
        super().__init__(device, n_classes)

        self.mse = MeanSquaredError(squared=True)
        self.rmse = MeanSquaredError(squared=False)
        self.scc = SpearmanCorrCoef()

    def compute_metrics(
            self, predicted: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None) -> Dict[str, Union[int, float]]:
        return {
            'mse': self._compute_metric(self.mse, predicted, labels).item(),
            'rmse': self._compute_metric(self.rmse, predicted, labels).item(),
            'spearmans-corr-coeff': self._compute_metric(self.scc, predicted, labels).item()
        }


class SequenceClassificationMetricsCalculator(ClassificationMetricsCalculator):
    pass


class ResidueClassificationMetricsCalculator(ClassificationMetricsCalculator):
    def compute_metrics(
            self, predicted: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None) -> Dict[str, Union[int, float]]:
        if predicted is not None and labels is not None:
            # This will flatten everything!
            masks = labels != MASK_AND_LABELS_PAD_VALUE
            masks = masks.to(self.device)

            masked_predicted = torch.masked_select(predicted, masks)
            masked_labels = torch.masked_select(labels, masks)

            return super().compute_metrics(predicted=masked_predicted, labels=masked_labels)
        else:
            return super().compute_metrics(predicted=predicted, labels=labels)

class ResiduesClassificationMetricsCalculator(ClassificationMetricsCalculator):
    pass

class SequenceRegressionMetricsCalculator(RegressionMetricsCalculator):
    pass

class ResidueRegressionMetricsCalculator(RegressionMetricsCalculator):
    def compute_metrics(
            self, predicted: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None) -> Dict[str, Union[int, float]]:
        predicted_flattened = predicted.flatten() if predicted is not None else None
        labels_flattened = labels.flatten() if labels is not None else None
        return {
            'mse': self._compute_metric(self.mse, predicted, labels).item(),
            'rmse': self._compute_metric(self.rmse, predicted, labels).item(),
            'spearmans-corr-coeff': self._compute_metric(self.scc, predicted_flattened, labels_flattened).item()
        }


class ResiduesRegressionMetricsCalculator(RegressionMetricsCalculator):
    pass



import re

from typing import Iterable, List, Optional


def preprocess_sequences_with_whitespaces(sequences: Iterable[str], mask_token: Optional[str]) -> List[str]:
    # Remove rare amino acids
    sequences_cleaned = [re.sub(r"[UZOB]", "X", sequence) for sequence in sequences]
    # Transformers need spaces between the amino acids

    def add_whitespaces(sequence: str) -> str:
        """ Do not split mask token """
        result = ""
        sequence_parts = sequence.split(sep=mask_token)
        for sequence_part in sequence_parts:
            if len(result) > 0:
                result += " "

            result += " ".join(list(sequence_part))
            result += f" {mask_token}"
        result = result[:-(len(mask_token)+1)].strip()  # +1 for last whitespace
        return result

    if mask_token is not None:
        sequences_with_spaces = [add_whitespaces(sequence) for sequence in sequences_cleaned]
    else:
        sequences_with_spaces = [" ".join(list(sequence)) for sequence in sequences_cleaned]
    return sequences_with_spaces


def preprocess_sequences_without_whitespaces(sequences: Iterable[str], mask_token: Optional[str]) -> List[str]:
    # Remove rare amino acids
    sequences_cleaned = [re.sub(r"[UZOB]", "X", sequence) for sequence in sequences]
    return sequences_cleaned


def preprocess_sequences_for_prostt5(sequences: Iterable[str], mask_token: Optional[str]) -> List[str]:
    preprocessed_sequences = preprocess_sequences_with_whitespaces(sequences, mask_token=mask_token)

    # We have AAs, so we need to tell the model we want to translate AA->3Di.
    # To do so, we prepend "<AA2fold>" while keeping residues in upper-case.
    prefixed_sequences = [
        "<AA2fold>" + " " + seq.upper() for seq in preprocessed_sequences
    ]
    return prefixed_sequences



from .random_embedder import RandomEmbedder
from .blosum62_embedder import Blosum62Embedder
from .aa_ontology_embedder import AAOntologyEmbedder
from .one_hot_encoding_embedder import OneHotEncodingEmbedder

__all__ = ["AAOntologyEmbedder", "OneHotEncodingEmbedder", "RandomEmbedder", "Blosum62Embedder"]



from typing import Optional, Any, Dict

from .pipeline import Pipeline, PipelineStep
from .steps import SetupStep, EmbeddingStep, FineTuningEmbeddingStep, ProjectionStep, DataLoadingStep, TrainingStep, \
    TestingStep, PostProcessStep, InputValidationStep


class DefaultPipeline:
    def __init__(self, config: Dict[str, Any]):
        if "finetuning_config" in config:
            self.pipeline = self._default_finetuning_pipeline()
        else:
            self.pipeline = self._default_pipeline()

    @staticmethod
    def _default_pipeline():
        return (Pipeline()
                .add_step(SetupStep())
                .add_step(InputValidationStep())
                .add_step(EmbeddingStep())
                .add_step(ProjectionStep())
                .add_step(DataLoadingStep())
                .add_step(TrainingStep())
                .add_step(TestingStep())
                .add_step(PostProcessStep())
                )

    @staticmethod
    def _default_finetuning_pipeline():
        return (Pipeline()
                .add_step(SetupStep())
                .add_step(InputValidationStep())
                .add_step(FineTuningEmbeddingStep())
                .add_step(DataLoadingStep())
                .add_step(TrainingStep())
                .add_step(TestingStep())
                .add_step(PostProcessStep())
                )

    def with_custom_steps(self,
                          custom_setup_step: Optional[PipelineStep] = None,
                          custom_input_validation_step: Optional[PipelineStep] = None,
                          custom_embedding_step: Optional[PipelineStep] = None,
                          custom_projection_step: Optional[PipelineStep] = None,
                          custom_data_loading_step: Optional[PipelineStep] = None,
                          custom_training_step: Optional[PipelineStep] = None,
                          custom_testing_step: Optional[PipelineStep] = None,
                          custom_postprocess_step: Optional[PipelineStep] = None):

        custom_steps = [
            custom_setup_step,
            custom_input_validation_step,
            custom_embedding_step,
            custom_projection_step,
            custom_data_loading_step,
            custom_training_step,
            custom_testing_step,
            custom_postprocess_step
        ]

        for custom_step in custom_steps:
            if custom_step is not None:
                self.pipeline.replace_step(custom_step)

        return self



import os

from ruamel import yaml
from pathlib import Path
from itertools import chain
from ruamel.yaml import YAMLError
from typing import Union, List, Dict, Any, Tuple

from .config_option import ConfigOption, ConfigKey
from .deprecated_config import deprecated_config_keys
from .general_config import general_config
from .input_config import input_config
from .model_config import model_config
from .training_config import training_config
from .embedding_config import embedding_config
from .finetuning_config import finetuning_config
from .hf_dataset_config import hf_dataset_config
from .config_exception import ConfigurationException
from .cross_validation_config import cross_validation_config, get_default_cross_validation_config
from .config_validation import validate_config_rules, validate_config_options
from .config_utils import download_file_from_config, is_url, make_path_absolute_if_necessary

from ..protocols import Protocol
from ..input_files import process_hf_dataset_to_fasta


class Configurator:
    """
    Class to read, validate, and transform the input YAML configuration.

    The `Configurator` class handles the parsing of configuration files, applies validation rules,
    transforms configuration values (such as downloading necessary files), and ensures that the
    configuration is consistent and adheres to the defined rules based on the selected protocol.
    """

    def __init__(
            self,
            config_dict: Dict,
            config_file_path: Path = None,
            allow_downloads: bool = True,
    ):
        """
        Initialize a Configurator instance.

        Args:
            config_dict (Dict): The configuration dictionary parsed from the YAML file.
            config_file_path (Path, optional): Path to the configuration file. Defaults to the current directory.
            allow_downloads (bool, optional): Whether to allow downloading files. Defaults to True.
        """
        if not config_file_path:
            config_file_path = Path("")
        self._config_file_path = config_file_path
        self._config_dict = config_dict
        self.allow_downloads = allow_downloads
        self.protocol = self._get_protocol_from_config_dict(config_dict)

    @classmethod
    def from_config_dict(cls, config_dict: Dict[str, Any]):
        """
        Create a Configurator instance from a configuration dictionary.

        Args:
            config_dict (Dict[str, Any]): The configuration dictionary.

        Returns:
            Configurator: An instance of the Configurator class initialized with the provided dictionary.
        """
        return cls(config_dict=config_dict)

    @classmethod
    def from_config_path(cls, config_path: Union[str, Path]):
        """
        Create a Configurator instance by reading a configuration file.

        Args:
            config_path (Union[str, Path]): Path to the YAML configuration file.

        Returns:
            Configurator: An instance of the Configurator class initialized with the configuration file.
        """
        return cls(
            config_dict=cls._read_config_file(config_path),
            config_file_path=Path(os.path.dirname(os.path.abspath(config_path))),
        )

    @staticmethod
    def get_option_dicts_by_protocol(
            protocol: Protocol,
            sub_configs_to_include: List[ConfigKey] = None
    ) -> List[Dict[str, Any]]:
        """
        Returns all possible configuration options as dictionaries for the given protocol.

        Args:
            protocol (Protocol): The protocol to get all options for.
            sub_configs_to_include (List[ConfigKey]): List of sub-configuration keys to include.
        Returns:
            List of config options as dictionaries.
        """
        result = []
        main_config, sub_configs = Configurator._get_relevant_config_options(protocol=protocol,
                                                                             config_keys=[str(config_key.value) for
                                                                                          config_key in
                                                                                          sub_configs_to_include])
        all_config_options = {**main_config,
                              **{k: v for sub_config in sub_configs.values() for k, v in sub_config.items()}}
        for option in all_config_options.values():
            if option.constraints and protocol not in option.constraints.allowed_protocols:
                continue
            result.append(option.to_dict())
        return result

    @staticmethod
    def _read_config_file(
            config_path: Union[str, Path],
            preserve_order: bool = True
    ) -> dict:
        """
        Read configuration from a YAML file.

        Args:
            config_path (Union[str, Path]): Path to the YAML configuration file.
            preserve_order (bool, optional): Whether to preserve the order of the YAML file. Defaults to True.

        Returns:
            dict: The configuration file parsed as a dictionary.

        Raises:
            ConfigurationException: If the YAML file cannot be parsed.
        """
        with open(config_path, "r") as fp:
            try:
                if preserve_order:
                    return yaml.load(fp, Loader=yaml.Loader)
                else:
                    return yaml.safe_load(fp)
            except YAMLError as e:
                raise ConfigurationException(
                    f"Could not parse configuration file at '{config_path}' as YAML. "
                    "Formatting mistake in config file? "
                    "See error above for details."
                ) from e

    @staticmethod
    def _get_protocol_from_config_dict(config_dict: Dict[str, Any]):
        """
        Extract the protocol from the configuration dictionary.

        Args:
            config_dict (Dict[str, Any]): The configuration dictionary.

        Returns:
            Protocol: The extracted protocol.

        Raises:
            ConfigurationException: If the protocol is not specified or invalid.
        """
        protocol = config_dict.get("protocol")
        if protocol is None:
            raise ConfigurationException(
                "No protocol specified in config file!"
            )
        try:
            return Protocol[protocol]
        except KeyError:
            raise ConfigurationException(
                f"Invalid protocol specified: {protocol}"
            )

    @staticmethod
    def _get_relevant_config_options(protocol: Protocol,
                                     config_keys: List[str]) -> (
            Tuple)[Dict[str, ConfigOption], Dict[ConfigKey, Dict[str, ConfigOption]]]:
        main_config = {config_option.name: config_option for config_option in chain(
            general_config(protocol)[1],
            input_config(protocol)[1],
            model_config(protocol)[1],
            training_config(protocol)[1],
            embedding_config(protocol)[1],
        )}
        sub_configs = {config_key: {config_option.name: config_option for config_option in config_options} for
                       config_key, config_options in
                       [hf_dataset_config(protocol), cross_validation_config(protocol), finetuning_config(protocol)]
                       if config_key.value in config_keys}

        return main_config, sub_configs

    def verify_config(self, ignore_file_checks: bool = False) -> Dict[str, Any]:
        """
        Reads the YAML configuration, performs value transformations (such as downloading files),
        and verifies the configuration's correctness.

        Args:
            ignore_file_checks (bool, optional): If True, file-related checks are not performed. Defaults to False.

        Returns:
            Dict[str, Any]: Dictionary with configuration option names as keys and their respective (transformed) values.

        Raises:
            ConfigurationException: If any validation rule is violated or if required options are missing.
        """
        if ConfigKey.CROSS_VALIDATION.value not in self._config_dict:
            self._config_dict.update(get_default_cross_validation_config())

        # Check config is deprecated
        deprecated_keys_with_replacement = deprecated_config_keys(config_dict=self._config_dict)
        if len(deprecated_keys_with_replacement) > 0:
            deprecation_information = "\n".join([f'Deprecated Key: {key} - Replacement: {replacement}'
                                                 for key, replacement
                                                 in deprecated_keys_with_replacement.items()])
            raise ConfigurationException(f"Config contains deprecated keys:\n{deprecation_information}")

        if not validate_config_rules(protocol=self.protocol,
                                     ignore_file_checks=ignore_file_checks,
                                     config_dict=self._config_dict):
            raise ConfigurationException(f"Provided config is not valid!")

        verified_config = {}
        main_config, sub_configs = self._get_relevant_config_options(protocol=self.protocol,
                                                                     config_keys=[config_key.value for config_key in
                                                                                  ConfigKey.all_subconfig_keys()
                                                                                  if config_key.value in
                                                                                  self._config_dict])
        verified_main_config = validate_config_options(protocol=self.protocol,
                                                       allow_downloads=self.allow_downloads,
                                                       ignore_file_checks=ignore_file_checks,
                                                       config_options=main_config,
                                                       config_dict=self._config_dict,
                                                       config_key=ConfigKey.ROOT)
        verified_config.update(verified_main_config)

        for config_key, config_options in sub_configs.items():
            verified_sub_config = validate_config_options(protocol=self.protocol,
                                                          allow_downloads=self.allow_downloads,
                                                          ignore_file_checks=ignore_file_checks,
                                                          config_options=config_options,
                                                          config_dict=self._config_dict[config_key.value],
                                                          config_key=config_key)
            verified_config[config_key.value] = verified_sub_config
        return verified_config

    def postprocess_config(self, verified_config: Dict[str, Any]) -> Dict[str, Any]:
        main_config, sub_configs = self._get_relevant_config_options(protocol=self.protocol,
                                                                     config_keys=[config_key.value for config_key in
                                                                                  ConfigKey.all_subconfig_keys() if
                                                                                  config_key.value in verified_config])
        all_config_options = {**main_config,
                              **{k: v for sub_config in sub_configs.values() for k, v in sub_config.items()}}

        # Output dir
        if "output_dir" in verified_config:
            output_dir = self._config_file_path / Path(verified_config["output_dir"])
            output_dir.mkdir(parents=True, exist_ok=True)
            verified_config["output_dir"] = make_path_absolute_if_necessary(value=verified_config["output_dir"],
                                                                            config_file_path=self._config_file_path)
        else:
            raise ConfigurationException("Verified config is missing output_dir option!")
        output_dir = Path(verified_config["output_dir"])

        if ConfigKey.HF_DATASET.value in verified_config:
            try:
                hf_dataset_dir = output_dir / "hf_db"
                hf_dataset_dir.mkdir(exist_ok=True)
                input_file_path = process_hf_dataset_to_fasta(
                    hf_storage_path=hf_dataset_dir,
                    hf_map=verified_config["hf_dataset"])
                if input_file_path:
                    verified_config["input_file"] = str(input_file_path)
            except Exception as e:
                raise ConfigurationException(f"Error while creating huggingface dataset files: {e}")

        postprocessed_config = {}
        for option_key, option_value in verified_config.items():
            try:
                if isinstance(option_value, dict):
                    postprocessed_config[option_key] = option_value
                    continue  # Nothing to postprocess in sub-configs yet

                config_option: ConfigOption = all_config_options[option_key]
            except KeyError:
                raise ConfigurationException(f"Invalid option key '{option_key}' on verified config.\n"
                                             f"Has the configuration already been verified?")
            # TODO [Refactoring] Special case for ONNX as embedder name - might need to replace file
            #  option with function to determine dynamically
            if (config_option.is_file_option or
                    (config_option.name == "embedder_name" and str(option_value).endswith(".onnx"))):
                if is_url(option_value):
                    if not self.allow_downloads:
                        raise ConfigurationException(f"Downloading files is disabled!")

                    option_value = download_file_from_config(option_name=option_key,
                                                             url=option_value,
                                                             config_file_path=self._config_file_path)
                option_value = make_path_absolute_if_necessary(value=option_value,
                                                               config_file_path=self._config_file_path)

            postprocessed_config[option_key] = option_value

        postprocessed_config["protocol"] = self.protocol.name
        return postprocessed_config

    def get_verified_config(self, ignore_file_checks: bool = False) -> Dict[str, Any]:
        """
        Reads the YAML configuration, performs value transformations (such as downloading files),
        and verifies the configuration's correctness.
        Convenience function to perform validation and postprocessing at once for backwards compatibility.

        Args:
            ignore_file_checks (bool, optional): If True, file-related checks are not performed. Defaults to False
        Returns:
            Dict[str, Any]: Dictionary with configuration option names as keys and their respective (transformed) values
        Raises:
            ConfigurationException: If any validation rule is violated or if required options are missing.
        """
        verified_config = self.verify_config(ignore_file_checks=ignore_file_checks)
        return self.postprocess_config(verified_config)



from pathlib import Path
from typing import Union, Dict, Any, Optional, List

from .logging import clear_logging

from ..config import Configurator
from ..trainers import Trainer, Pipeline
from ..output_files import OutputManager, output_observer_factory, BiotrainerOutputObserver


def parse_config_file_and_execute_run(config: Union[str, Path, Dict[str, Any]],
                                      custom_pipeline: Optional[Pipeline] = None,
                                      custom_output_observers: Optional[List[BiotrainerOutputObserver]] = None) \
        -> Dict[str, Any]:
    # Verify config via configurator
    configurator = None
    if isinstance(config, str):
        configurator = Configurator.from_config_path(config)
    elif isinstance(config, Path):
        configurator = Configurator.from_config_path(str(config))
    elif isinstance(config, dict):
        configurator = Configurator.from_config_dict(config)

    assert configurator is not None, f"Config could not be read, incorrect type: {type(config)}"

    config = configurator.get_verified_config(ignore_file_checks=False)
    output_dir = Path(config["output_dir"])

    output_observers = output_observer_factory(output_dir=output_dir, config=config)
    if custom_output_observers and len(custom_output_observers) > 0:
        output_observers.extend(custom_output_observers)

    output_manager = OutputManager(observers=output_observers)
    output_manager.add_config(config=config)

    trainer: Trainer
    if custom_pipeline:
        output_manager.add_derived_values(derived_values={"custom_pipeline": True})

    # Run biotrainer pipeline
    trainer = Trainer(config=config,
                      output_manager=output_manager,
                      custom_pipeline=custom_pipeline
                      )

    output_manager = trainer.run()

    # Save output_variables in out.yml
    output_result = output_manager.write_to_file(output_dir=output_dir)

    clear_logging()

    return output_result



from __future__ import annotations

import torch

from enum import Enum
from typing import List


class Protocol(Enum):
    residue_to_value = 0,
    residue_to_class = 1,
    residues_to_class = 2,
    residues_to_value = 3,
    sequence_to_class = 4,
    sequence_to_value = 5

    @staticmethod
    def all() -> List[Protocol]:
        return [Protocol.residue_to_class,
                Protocol.residue_to_value,
                Protocol.residues_to_class,
                Protocol.residues_to_value,
                Protocol.sequence_to_class,
                Protocol.sequence_to_value]

    @staticmethod
    def classification_protocols() -> List[Protocol]:
        return [Protocol.residue_to_class, Protocol.residues_to_class, Protocol.sequence_to_class]

    @staticmethod
    def regression_protocols() -> List[Protocol]:
        return [Protocol.residue_to_value, Protocol.residues_to_value, Protocol.sequence_to_value]

    @staticmethod
    def per_sequence_protocols() -> List[Protocol]:
        return [Protocol.residues_to_class, Protocol.residues_to_value,
                Protocol.sequence_to_class, Protocol.sequence_to_value]

    @staticmethod
    def per_residue_protocols() -> List[Protocol]:
        return [Protocol.residue_to_class, Protocol.residue_to_value]

    @staticmethod
    def using_per_residue_embeddings() -> List[Protocol]:
        return [Protocol.residue_to_class, Protocol.residue_to_value,
                Protocol.residues_to_class, Protocol.residues_to_value]

    @staticmethod
    def using_per_sequence_embeddings() -> List[Protocol]:
        return [Protocol.sequence_to_class, Protocol.sequence_to_value]

    @staticmethod
    def from_string(string: str) -> Protocol:
        return {p.name: p for p in Protocol.all()}[string]

    def get_dummy_input(self, embedding_dimension: int):
        batch_size = 1
        default_sequence_length = 50
        if self in Protocol.using_per_residue_embeddings():
            return torch.rand((batch_size, default_sequence_length, embedding_dimension), dtype=torch.float32)
        return torch.rand((batch_size, embedding_dimension), dtype=torch.float32)

    def __str__(self):
        return self.name



import unittest

from biotrainer.embedders.interfaces import preprocess_sequences_with_whitespaces, preprocess_sequences_without_whitespaces


class EmbeddingUtilsTests(unittest.TestCase):

    def test_preprocessing_whitespace(self):
        mask_tokens = ["[MASK]", "<extra_id_0>"]

        # Basic cases without tokens
        sequences_without_token = [
            "PRTEINS",
            "S",  # Single amino acid
            "QESWENSCE",
            "SEQS",
            "UZOB",  # Special amino acids that should be converted to X
            "PROTEIN" * 10,  # Long sequence
        ]

        # Edge cases with tokens
        edge_cases = [
            "{token}",  # Only token
            "{token}{token}",  # Multiple consecutive tokens
            "A{token}",  # Token at end
            "{token}A",  # Token at start
            "A{token}A{token}A",  # Multiple tokens with single amino acids
            " {token} ",  # Spaces around token (should be cleaned)
            "UZO{token}B",  # Special amino acids with token
        ]

        # Test sequences without tokens
        for mask_token in mask_tokens:
            preprocessed = preprocess_sequences_with_whitespaces(sequences_without_token, mask_token)
            preprocessed_with_none = preprocess_sequences_with_whitespaces(sequences_without_token, None)

            for sequence in preprocessed + preprocessed_with_none:
                # Basic checks
                self.assertTrue(sequence[-1] != " ")
                self.assertTrue(mask_token not in sequence)

                # Check spacing
                self.assertEqual(len(sequence.split()), len(sequence.replace(" ", "")))

                # Check special amino acid conversion
                self.assertFalse(any(aa in sequence for aa in "UZOB"))

                # Check no double spaces
                self.assertFalse("  " in sequence)

        # Test sequences with tokens
        for mask_token in mask_tokens:
            # Format edge cases with actual token
            test_sequences = [case.format(token=mask_token) for case in edge_cases]

            # Add more complex cases
            test_sequences.extend([
                f"PROTEIN{mask_token}SEQUENCE",
                f"{mask_token}".join(["A"] * 5),  # Multiple tokens
                f"PRO{mask_token}TE{mask_token}IN",
                f"{mask_token}PROTEIN{mask_token}",
                f"A{mask_token}" * 5,  # Repeating pattern
            ])

            preprocessed = preprocess_sequences_with_whitespaces(test_sequences, mask_token)

            for orig, processed in zip(test_sequences, preprocessed):
                # Token preservation
                self.assertEqual(
                    orig.count(mask_token),
                    processed.count(mask_token),
                    f"Token count mismatch in {orig} -> {processed}"
                )

                # Check token integrity
                token_parts = processed.split(mask_token)
                for part in token_parts:
                    if part:  # Skip empty parts
                        # Each amino acid should be space-separated
                        self.assertEqual(
                            len(part.strip().split()),
                            len(part.strip().replace(" ", "")),
                            f"Incorrect spacing in part: {part}"
                        )

                # Check special amino acid conversion
                self.assertFalse(any(aa in processed for aa in "UZOB"))

                # Check no trailing/leading spaces
                self.assertEqual(processed, processed.strip())

        # Check that tokens in the middle of the sequence are correctly handled e.g. T[MASK]P -> T [MASK] P
        for mask_token in mask_tokens:
            preprocessed = preprocess_sequences_with_whitespaces([f"T{mask_token}P"], mask_token=mask_token)
            for processed in preprocessed:
                self.assertTrue(f" {mask_token} " in processed)


    def test_preprocessing_without_whitespace(self):
        sequences_without_token = [
            "PRTEINS",
            "S",  # Single amino acid
            "QESWENSCE",
            "SEQS",
            "UZOB",  # Special amino acids that should be converted to X
            "PROTEIN" * 10,  # Long sequence
        ]
        preprocessed = preprocess_sequences_without_whitespaces(sequences_without_token, mask_token=None)

        for processed in preprocessed:
            self.assertFalse(" " in processed)
            self.assertEqual(processed, processed.strip())
            self.assertFalse(any(aa in processed for aa in "UZOB"))



import gc

from pathlib import Path

from ..pipeline import PipelineContext, PipelineStep
from ..pipeline.pipeline_step import PipelineStepType

from ...utilities import get_logger
from ...embedders import get_embedding_service, EmbeddingService

logger = get_logger(__name__)


class EmbeddingStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.EMBEDDING

    def process(self, context: PipelineContext) -> PipelineContext:
        # Generate embeddings if necessary, otherwise use existing embeddings
        embeddings_file = context.config.get("embeddings_file", None)

        if not embeddings_file:
            # Search for embeddings file at default place if no custom file was provided directly
            embeddings_file = EmbeddingService.get_embeddings_file_path(output_dir=context.config["output_dir"],
                                                                        protocol=context.config["protocol"],
                                                                        embedder_name=context.config["embedder_name"],
                                                                        use_half_precision=context.config.get(
                                                                            "use_half_precision"),
                                                                        )

        if not embeddings_file or not Path(embeddings_file).is_file():
            embedding_service: EmbeddingService = get_embedding_service(
                custom_tokenizer_config=context.config.get("custom_tokenizer_config"),
                embedder_name=context.config["embedder_name"],
                use_half_precision=context.config.get("use_half_precision"),
                device=context.config["device"]
            )
            embeddings_file = embedding_service.compute_embeddings(
                input_data=context.input_data,
                protocol=context.config["protocol"], output_dir=context.config["output_dir"]
            )

            # Manually clear the memory from costly embedder model
            del embedding_service._embedder
            gc.collect()
        else:
            logger.info(f'Embeddings file was found at {embeddings_file}. Embeddings have not been computed.')

        context.output_manager.add_derived_values({'embeddings_file': str(embeddings_file)})

        # Mapping from id to embeddings
        id2emb = EmbeddingService.load_embeddings(embeddings_file_path=str(embeddings_file))

        context.id2emb = id2emb
        return context


class FineTuningEmbeddingStep(EmbeddingStep):
    def process(self, context: PipelineContext) -> PipelineContext:
        # Calculate / Load embeddings from not-finetuned model once in the beginning for baselines later
        logger.info(f'Finetuning embedder model is activated. Non-finetuned embeddings are calculated once at the '
                    f'beginning to calculate baselines later.')
        context = super().process(context)

        embedding_service = get_embedding_service(
            custom_tokenizer_config=context.config.get("custom_tokenizer_config"),
            embedder_name=context.config["embedder_name"],
            use_half_precision=context.config.get("use_half_precision"),
            device=context.config["device"],
            finetuning_config=context.config["finetuning_config"]
        )

        context.embedding_service = embedding_service
        return context



from ..pipeline import PipelineStep, PipelineContext
from ..pipeline.pipeline_step import PipelineStepType

from ...validations import InputValidator


class InputValidationStep(PipelineStep):
    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.INPUT_VALIDATION

    def process(self, context: PipelineContext) -> PipelineContext:
        if context.config.get("validate_input", True):
            protocol = context.config["protocol"]
            input_validator = InputValidator(protocol=protocol)
            validated_input_data = input_validator.validate(context.input_data)
            # No errors - set validated input data as input data
            context.input_data = validated_input_data
        return context


import torch
import inspect

from torch import nn
from typing import Dict, Set, Any, Optional

from .cnn import CNN
from .log_reg import LogReg
from .fnn import FNN, DeeperFNN
from .model_params import count_parameters
from .light_attention import LightAttention
from .biotrainer_model import BiotrainerModel
from .fine_tuning_model import FineTuningModel

from ..protocols import Protocol
from ..utilities import get_logger

logger = get_logger(__name__)

__MODELS = {
    Protocol.residue_to_class: {
        'CNN': CNN,
        'FNN': FNN,
        'DeeperFNN': DeeperFNN,
        'LogReg': LogReg,
    },
    Protocol.residue_to_value: {
        'CNN': CNN,
        'FNN': FNN,
        'DeeperFNN': DeeperFNN,
        'LogReg': LogReg,
    },
    Protocol.residues_to_class: {
        'LightAttention': LightAttention,
    },
    Protocol.residues_to_value: {
        'LightAttention': LightAttention,
    },
    Protocol.sequence_to_class: {
        'FNN': FNN,
        'DeeperFNN': DeeperFNN,
        'LogReg': LogReg
    },
    Protocol.sequence_to_value: {
        'FNN': FNN,
        'DeeperFNN': DeeperFNN,
        'LogReg': LogReg
    },
}

def _initialize_weights(module, init_type='default'):
    """
    Initialize weights for neural network modules

    Args:
        module: PyTorch module
        init_type: Type of initialization ('default', 'normal', 'xavier', 'kaiming')
    """
    if init_type == 'default':
        return

    if isinstance(module, (nn.Linear, nn.Conv1d, nn.Conv2d)):
        if init_type == 'normal':
            nn.init.normal_(module.weight, mean=0.0, std=0.01)
        elif init_type == 'xavier':
            nn.init.xavier_uniform_(module.weight)
        elif init_type == 'kaiming':
            nn.init.kaiming_uniform_(module.weight, nonlinearity='relu')
        # Default uses PyTorch's default initialization

        if init_type != 'default':
            nn.init.zeros_(module.bias)


def get_model(protocol: Protocol, model_choice: str, n_classes: int, n_features: int,
              disable_pytorch_compile: Optional[bool] = True,
              model_weights_init: Optional[str] = 'default',
              **kwargs):
    model_class = __MODELS.get(protocol).get(model_choice)
    if not model_class:
        raise NotImplementedError
    else:
        if "dropout_rate" in kwargs.keys() and "dropout_rate" not in inspect.signature(model_class).parameters:
            logger.warning(f"dropout_rate not implemented for model_choice {model_choice}")

        model = model_class(n_classes=n_classes, n_features=n_features, **kwargs)

        # Apply weight initialization if specified
        if model_weights_init != 'default':
            model.apply(lambda m: _initialize_weights(m,
                                                     init_type=model_weights_init,
                                                     ))

        # Disable option for backwards compatibility with older models or if there emerge problems during training
        if not disable_pytorch_compile:
            logger.info(f"Using pytorch model compile mode!")
            model.compile()
        return model


def get_available_models_dict() -> Dict[Protocol, Dict[str, Any]]:
    return dict(__MODELS)


def get_available_models_set() -> Set[str]:
    all_models_by_protocol = [list(protocol.keys()) for protocol in get_available_models_dict().values()]
    all_models_list = [model for model_list in all_models_by_protocol for model in model_list]
    return set(all_models_list)


__all__ = [
    "get_model",
    "get_available_models_dict",
    "get_available_models_set",
    "count_parameters",
    "BiotrainerModel",
    "FineTuningModel",
]



import random
import itertools

from typing import Dict, Any, Generator, Union

from ..utilities import get_logger
from ..config import ConfigurationException

logger = get_logger(__name__)


class HyperParameterManager:
    def __init__(self, **kwargs):
        self._constant_params = {}
        self._params_to_optimize = {}
        for key, value in kwargs.items():
            if "range" in str(value):
                try:
                    compiled_range = compile(source=value, filename='<range>', mode='eval')
                    range_obj = eval(compiled_range, locals())
                    assert range_obj, "Range could not be compiled from input!"
                    self._params_to_optimize[key] = range_obj
                except Exception as e:
                    raise ConfigurationException(f"Unable to compile hyper_parameters from config:\n"
                                                 f"{key}: {value}") from e
            elif type(value) is list:
                self._params_to_optimize[key] = list(set(value))
            elif type(value) is str and "[" in value and "]" in value:
                try:
                    compiled_list = compile(source=value, filename='<list>', mode='eval')
                    list_obj = eval(compiled_list, locals())
                    assert list_obj, "List could not be compiled from input!"
                    self._params_to_optimize[key] = list_obj
                except Exception as e:
                    raise ConfigurationException(f"Unable to compile hyper_parameters from config:\n"
                                                 f"{key}: {value}") from e
            else:
                self._constant_params[key] = value

        if "nested" in self._constant_params["cross_validation_config"].keys() and \
                str(self._constant_params["cross_validation_config"]["nested"]).capitalize() == "True":
            if len(self._params_to_optimize.keys()) == 0:
                raise ConfigurationException(f"No parameters to optimize were given. Define with list or range!\n"
                                             f"e.g. learning_rate: [1e-3, 1e-4]")
        if len(self._params_to_optimize.keys()) > 0 and \
                self._constant_params["cross_validation_config"]["method"] != "k_fold":
            raise ConfigurationException(f"Parameter search only supported for nested k_fold cross validation!")

    def search(self, mode="no_search") -> \
            Generator[Dict[str, Any], None, Union[None, Generator[Dict[str, Any], None, None]]]:
        if mode == "no_search":
            return self._no_search()
        elif mode == "grid_search":
            return self._grid_search()
        elif mode == "random_search":
            return self._random_search()
        else:
            raise NotImplementedError(f"Hyper parameter search method {mode} not implemented!")

    def get_only_params_to_optimize(self, hyper_params: Dict[str, Any]) -> Dict[str, Any]:
        hp_to_optimize = {}
        for key, value in hyper_params.items():
            if key in self._params_to_optimize.keys():
                hp_to_optimize[key] = value
        return hp_to_optimize

    def _no_search(self) -> Generator[Dict[str, Any], None, None]:
        yield self._constant_params

    def _grid_search(self) -> Generator[Dict[str, Any], None, None]:
        keys, values = zip(*self._params_to_optimize.items())

        for combination in itertools.product(*values):
            hyper_parameters = dict(zip(keys, combination))
            yield {**self._constant_params, **hyper_parameters}

    def _random_search(self) -> Generator[Dict[str, Any], None, None]:
        n_max_evaluations = self._constant_params["cross_validation_config"]["n_max_evaluations_random"]
        all_possible_hp_combinations = list(self._grid_search())
        if n_max_evaluations >= len(all_possible_hp_combinations):
            logger.warning(f"Number of max_evaluations ({n_max_evaluations}) >= "
                           f"Number of possible combinations ({len(all_possible_hp_combinations)}) for random search "
                           f"hyperparameter optimization! Performing grid search instead..")
            yield from self._grid_search()

        random.shuffle(all_possible_hp_combinations)

        for i in range(n_max_evaluations):
            yield {**self._constant_params, **all_possible_hp_combinations[i]}



class ConfigurationException(Exception):
    """
    Exception raised for invalid configurations.

    This exception is used to indicate errors related to configuration validation and processing.
    """



import unittest

from functools import reduce
from itertools import combinations
from biotrainer.trainers import hp_manager


class HyperParameterSearchTests(unittest.TestCase):
    def test_hps_grid(self):
        # Only normal lists
        param_dict = {
            "cross_validation_config": {
                "method": "k_fold",
                "search_method": "grid_search"
            },
            "optimizer_choice": "adam",
            "use_class_weights": [True, False],
            "batch_size": [16, 48, 128],
            "learning_rate": [1e-3, 1e-4, 1e-5]
        }
        hps = hp_manager.HyperParameterManager(**param_dict)
        number_all_combinations = reduce(lambda l1, l2: l1 * l2,
                                         map(len, [val for val in param_dict.values()
                                                   if val != "adam" and type(val) != dict]))
        self.assertTrue(len(list(hps._grid_search())) == number_all_combinations)

        # Range and list comprehension
        param_dict = {
            "cross_validation_config": {
                "method": "k_fold",
                "search_method": "grid_search"
            },
            "optimizer_choice": "adam",
            "use_class_weights": [True, False],
            "batch_size": "range(0, 10, 1)",
            "learning_rate": "[10**-x for x in [3,4,5]]"
        }
        hps = hp_manager.HyperParameterManager(**param_dict)
        number_all_combinations = reduce(lambda l1, l2: l1 * l2, map(len, [[True, False], list(range(0, 10, 1)),
                                                                           [10 ** -x for x in [3, 4, 5]]]))
        self.assertTrue(len(list(hps._grid_search())) == number_all_combinations)

    def test_hps_random(self):
        n_max_evaluations = 10
        param_dict = {
            "optimizer_choice": "adam",
            "cross_validation_config": {
                "method": "k_fold",
                "search_method": "random_search",
                "n_max_evaluations_random": n_max_evaluations
            },
            "use_class_weights": [True, False],
            "batch_size": "range(0, 10, 1)",
            "learning_rate": "[10**-x for x in [3,4,5]]"
        }
        hps = hp_manager.HyperParameterManager(**param_dict)
        hyper_params_returned = []
        for hyper_params in hps._random_search():
            hyper_params_returned.append(hyper_params)
        self.assertTrue(len(hyper_params_returned) <= n_max_evaluations,
                        "random_search did not return correct amount of combinations!")

        all_combinations = combinations(hyper_params_returned, 2)
        all_unequal = all(map(lambda dicts: dicts[0] != dicts[1], all_combinations))
        self.assertTrue(all_unequal, msg="Random search produces redundant hyper parameter combinations!")


if __name__ == '__main__':
    unittest.main()



from .custom_tokenizer import CustomTokenizer
from .embedder_interfaces import EmbedderInterface, EmbedderWithFallback
from .preprocessing_strategies import preprocess_sequences_without_whitespaces, preprocess_sequences_with_whitespaces, \
    preprocess_sequences_for_prostt5

__all__ = ["EmbedderInterface",
           "EmbedderWithFallback",
           "preprocess_sequences_without_whitespaces",
           "preprocess_sequences_with_whitespaces",
           "preprocess_sequences_for_prostt5",
           "CustomTokenizer"
           ]



import os
import shutil
import tempfile
import cyclopts

from pathlib import Path
from typing import Union, Dict, Any, Optional

from .hashing import calculate_sequence_hash
from .executer import parse_config_file_and_execute_run

from ..trainers import Pipeline
from ..inference import Inferencer
from ..autoeval import autoeval_pipeline
from ..embedders import get_embedding_service
from ..input_files import convert_deprecated_fastas, read_FASTA

app = cyclopts.App()


@app.command
def train(config: Union[str, Path, Dict[str, Any]]) -> Dict[str, Any]:
    """
       Entry point for training

       @param config: Biotrainer configuration file path or config dict
       """
    return parse_config_file_and_execute_run(config)


def train_with_custom_pipeline(config: Union[str, Path, Dict[str, Any]],
                               custom_pipeline: Pipeline) -> Dict[str, Any]:
    """
    Entry point for usage in scripts with a custom pipeline to be executed instead of the default biotrainer pipeline.
    This can, for example, be useful, if embeddings should be loaded from a database instead of a file.

    @param config: Biotrainer configuration file path or config dict
    @param custom_pipeline: custom pipeline to execute
    """
    return parse_config_file_and_execute_run(config, custom_pipeline=custom_pipeline)


@app.command
def predict(training_output_file: Union[str, Path], model_input: str,
            save_embeddings: Optional[bool] = False) -> Dict[str, Any]:
    if isinstance(model_input, str):
        if "." in model_input and Path(model_input).exists():
            model_input = read_FASTA(model_input)
            input_ids = {seq_record.get_hash(): seq_record.seq_id for seq_record in model_input}
            model_input = {seq_record.get_hash(): seq_record.seq for seq_record in model_input}
        else:
            model_input = [seq for seq in model_input.split(",")]
            input_ids = {calculate_sequence_hash(seq): f"Seq{idx}" for idx, seq in enumerate(model_input)}
    else:
        raise ValueError("model_input must be a Path to an input file or a comma separated list of sequences!")

    inferencer, iom = Inferencer.create_from_out_file(out_file_path=training_output_file,
                                                      automatic_path_correction=True)

    embedding_service = get_embedding_service(embedder_name=iom.embedder_name(),
                                              custom_tokenizer_config=None,  # TODO
                                              use_half_precision=iom.use_half_precision(),
                                              device=iom.device())
    adapter_path = iom.adapter_path()
    if adapter_path is not None:
        embedding_service.add_finetuned_adapter(adapter_path=adapter_path)

    with tempfile.TemporaryDirectory() as tmpdir:
        result_file = embedding_service.compute_embeddings(input_data=model_input,
                                                           output_dir=Path(tmpdir),
                                                           protocol=iom.protocol(),
                                                           )
        embeddings = embedding_service.load_embeddings(result_file)
        if save_embeddings:
            shutil.copy(result_file, os.getcwd())

    result = inferencer.from_embeddings(embeddings=embeddings)["mapped_predictions"]

    sorted_results = []
    for seq_hash, prediction in result.items():
        input_id = input_ids[seq_hash]
        sorted_results.append((input_id, seq_hash, prediction))

    sorted_results = sorted(sorted_results, key=lambda x: x[0])

    for input_id, seq_hash, prediction in sorted_results:
        print(f"Prediction for {input_id} (sequence hash {seq_hash}):\n\t{prediction}")

    return result


@app.command
def convert(sequence_file: str,
            labels_file: Optional[str] = None,
            masks_file: Optional[str] = None,
            converted_file: Optional[str] = "converted.fasta",
            skip_inconsistencies: Optional[bool] = False,
            target_format: Optional[str] = "fasta"):
    convert_deprecated_fastas(result_file=converted_file,
                              sequence_file=sequence_file,
                              labels_file=labels_file,
                              masks_file=masks_file,
                              skip_sequence_on_failed_merge=skip_inconsistencies)


@app.command
def autoeval(embedder_name: str,
             framework: str,
             min_seq_length: Optional[int] = 0,
             max_seq_length: Optional[int] = 2000,
             use_half_precision: Optional[bool] = False,
             ):
    for progress in autoeval_pipeline(embedder_name=embedder_name,
                                      framework=framework,
                                      min_seq_length=min_seq_length,
                                      max_seq_length=max_seq_length,
                                      use_half_precision=use_half_precision,
                                      ):
        print(progress)


if __name__ == "__main__":
    app()



import torch
import torch.nn as nn
import biotrainer.utilities as utils

from .biotrainer_model import BiotrainerModel


# Convolutional neural network (two convolutional layers)
class CNN(BiotrainerModel):
    def __init__(
            self, n_classes: int, n_features: int,
            bottleneck_dim: int = 32,
            dropout_rate: float = 0.0,  # Dropout for CNN disabled by default
            **kwargs,
    ):
        super(CNN, self).__init__()

        self.conv1 = nn.Conv2d(n_features, bottleneck_dim, kernel_size=(7, 1), padding=(3, 0))
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout_rate)
        self.conv2 = nn.Conv2d(bottleneck_dim, n_classes, kernel_size=(7, 1), padding=(3, 0))

    def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
            L = protein length
            B = batch-size
            F = number of features (1024 for embeddings)
            N = number of classes (9 for conservation)
        """

        # Calculate mask
        mask = (x.sum(dim=-1) != utils.SEQUENCE_PAD_VALUE).unsqueeze(1).unsqueeze(3)  # Shape: (B, 1, L, 1)

        x = x.permute(0, 2, 1).unsqueeze(3)  # Shape: (B, F, L, 1)

        # First convolution
        x = self.conv1(x)
        x = self.relu(x)
        x = x * mask  # Apply mask

        # Dropout
        x = self.dropout(x)

        # Second convolution
        x = self.conv2(x)
        x = x * mask  # Apply mask

        # Remove the last dimension and permute back
        x = x.squeeze(3).permute(0, 1, 2)  # Shape: (B, L, N)

        return x



import os
import psutil
import random
import unittest
import tempfile

from pathlib import Path

from biotrainer.input_files import BiotrainerSequenceRecord
from biotrainer.protocols import Protocol
from biotrainer.embedders import OneHotEncodingEmbedder, EmbeddingService


class TestEmbeddingService(unittest.TestCase):
    def setUp(self):
        self._setup_test_environment()
        self._setup_fasta_parameters()

    def _setup_test_environment(self):
        print('Setting up test environment')
        self.embedder = OneHotEncodingEmbedder()
        self.embedding_service = EmbeddingService(embedder=self.embedder)

    def _setup_fasta_parameters(self):
        self.num_reads = 100
        if False:
            # Test with large sequences disabled - It has been tested throughout and should only be activated again
            # if problems arise, because not doing this test hugely speeds up the CI pipeline
            print("Generating ultra-long sequences to test memory limits and extreme edge cases.")
            print(
                "Sequence generation may take considerable time due to large memory allocation: "
                f"{psutil.virtual_memory().available / (1024 ** 3):.2f} GB available."
            )
            print(
                "The calculations for embeddings of these ultra-long sequences are also computationally intensive, "
                "which contributes to the overall long duration of the test."
            )
            # Use the original calculation in CI environment
            self.long_length = int((0.75 * psutil.virtual_memory().available) / (18 * 21))
        else:
            # Use a fixed value for local development
            self.long_length = 50000
        self.other_length = 250

    @staticmethod
    def _generate_sequence(length):
        """Generate a random sequence of a given length."""
        return ''.join(random.choice("ACDE") for _ in range(length))

    def _generate_fasta(self, num_reads, long_length, other_length, filename, include_long=True, include_short=True):
        """Generate a FASTA file with a specified number of reads."""
        print(f"Generating FASTA file: {filename}")
        with open(filename, 'w') as file:
            if include_long:
                print(f"Generating long sequence, it may take a bit of time.")
                for i in range(1, 3):
                    file.write(f">read_{i}\n{self._generate_sequence(long_length)}\n")

            if include_short:
                for i in range(3, num_reads + 1):
                    file.write(f">read_{i}\n{self._generate_sequence(other_length)}\n")
        print(f"FASTA file generated: {filename}")

    def _run_embedding_test(self, test_name, num_reads, protocol, include_long=True, include_short=True):
        with tempfile.TemporaryDirectory() as tmp_dir:
            print(f'Starting test: compute_embeddings_{test_name}')
            sequence_path = os.path.join(tmp_dir, f"{test_name}.fasta")
            self._generate_fasta(num_reads, self.long_length, self.other_length, sequence_path, include_long,
                                 include_short)
            result = self._compute_embeddings(sequence_path, tmp_dir, protocol)
            self._verify_result(protocol, result, tmp_dir)
            print(f'Test compute_embeddings_{test_name} completed successfully')

    def _compute_embeddings(self, sequence_path, output_dir, protocol):
        print('Computing embeddings')
        return self.embedding_service.compute_embeddings(
            sequence_path,
            Path(output_dir),
            protocol
        )

    def _verify_result(self, protocol, result, tmp_dir):
        self.assertTrue(os.path.exists(result), f"Result file does not exist: {result}")
        if protocol == Protocol.sequence_to_class:
            expected_path = os.path.join(tmp_dir, "sequence_to_class", "one_hot_encoding",
                                         "reduced_embeddings_file_one_hot_encoding.h5")
        elif protocol == Protocol.residue_to_class:
            expected_path = os.path.join(tmp_dir, "residue_to_class", "one_hot_encoding",
                                         "embeddings_file_one_hot_encoding.h5")
        self.assertEqual(result, expected_path, f"Unexpected result path. Expected {expected_path}, got {result}")

    # Test methods
    def test_long_sequence_to_class(self):
        self._run_embedding_test("long_sequences", 2, Protocol.sequence_to_class, include_short=False)

    def test_short_sequence_to_class(self):
        self._run_embedding_test("short_sequences", self.num_reads, Protocol.sequence_to_class, include_long=False)

    def test_mixed_sequence_to_class(self):
        self._run_embedding_test("mixed_sequences", self.num_reads, Protocol.sequence_to_class)

    def test_long_residue_to_class(self):
        self._run_embedding_test("long_sequences", 2, Protocol.residue_to_class, include_short=False)

    def test_short_residue_to_class(self):
        self._run_embedding_test("short_sequences", self.num_reads, Protocol.residue_to_class, include_long=False)

    def test_mixed_residue_to_class(self):
        self._run_embedding_test("mixed_sequences", self.num_reads, Protocol.residue_to_class)

    def test_not_allowed_seq_ids(self):
        seq_records = [BiotrainerSequenceRecord(seq_id="not/Allowed", seq="MMAAAG"),
                       BiotrainerSequenceRecord(seq_id="ALLOWED", seq="MMAAGX")
                       ]
        with tempfile.TemporaryDirectory() as tmp_dir:
            with self.assertRaises(ValueError):
                self.embedding_service.compute_embeddings(input_data=seq_records,
                                                          output_dir=Path(tmp_dir),
                                                          protocol=Protocol.sequence_to_class,
                                                          store_by_hash=False)


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from collections import namedtuple

Split = namedtuple("Split", "name train val")
SplitResult = namedtuple("SplitResult", "name, hyper_params, best_epoch_metrics, solver")
EmbeddingDatasetSample = namedtuple("EmbeddingDatasetSample", "seq_id embedding target")
SequenceDatasetSample = namedtuple("SequenceDatasetSample", "seq_id, seq_record target")

@dataclass
class EpochMetrics:
    epoch: int
    training: dict
    validation: dict

    def to_dict(self):
        return {"epoch": self.epoch, "training": self.training, "validation": self.validation}



from __future__ import annotations

import os
import time
import h5py
import torch

from tqdm import tqdm
from pathlib import Path
from sklearn.manifold import TSNE
from typing import Dict, List, Union, Optional, Generator, Tuple, Any

from ..interfaces import EmbedderInterface

from ...protocols import Protocol
from ...utilities import get_logger, is_running_in_notebook
from ...input_files import read_FASTA, BiotrainerSequenceRecord

logger = get_logger(__name__)


class EmbeddingService:
    """
    A service class for computing embeddings using a provided embedder.
    """

    def __init__(self, embedder: EmbedderInterface = None, use_half_precision: bool = False):
        self._embedder = embedder
        self._use_half_precision = use_half_precision

    def save_embedder(self, output_dir: Path):
        """Save fine-tuned model after training"""
        logger.warning("Trying to save non-finetuned model!")

    def add_finetuned_adapter(self, adapter_path: Path) -> EmbeddingService:
        """Add finetuned adapter after training to embedder model. Used for inference."""
        from peft import PeftModel
        # Apply LoRA adapters to the embedder model
        model = self._embedder._model
        if model is None:
            raise ValueError(f"{self._embedder.name} does not provide a model to add the finetuning adapter!")

        model = PeftModel.from_pretrained(
            self._embedder._model,
            adapter_path
        )

        self._embedder._model = model.eval()

        print(f"Added fine-tuned adapter to {self._embedder.name}: {adapter_path}")
        return self

    def compute_embeddings(self,
                           input_data: Union[str, Path, List[str],
                           List[BiotrainerSequenceRecord], Dict[str, BiotrainerSequenceRecord]],
                           output_dir: Path,
                           protocol: Protocol,
                           force_output_dir: bool = False,
                           force_recomputing: bool = False,
                           store_by_hash: bool = True) -> str:
        """
        Compute embeddings with the provided embedder from a sequence file or a dictionary of sequences.

        Parameters:
            input_data (Union[str, Dict[str, str]]): Path to the sequence file or a dictionary of protein sequences.
            output_dir (Path): Output directory to store the computed embeddings.
            protocol (Protocol): Protocol for the embeddings. Determines if the embeddings should be reduced to per-protein.
            force_output_dir (bool): If True, the given output directory is directly used to store the embeddings file.
            force_recomputing (bool): If True, the embedding file is re-computed, even if it already exists.
            store_by_hash (bool): If True, sequence hashes are used as indices for the h5 result file.

        Returns:
            str: Path to the generated output h5 embeddings file.
        """
        use_reduced_embeddings = protocol in Protocol.using_per_sequence_embeddings()
        embeddings_file_path = self.get_embeddings_file_path(
            output_dir=output_dir,
            protocol=protocol,
            embedder_name=self._embedder.name,
            use_half_precision=self._use_half_precision,
            force_output_dir=force_output_dir
        )

        # Avoid re-computation if file already exists
        if not force_recomputing and embeddings_file_path.is_file():
            logger.info(f"Using existing embeddings file at {embeddings_file_path}")
            return str(embeddings_file_path)

        logger.info(f"Computing embeddings to: {str(embeddings_file_path)}")

        # Process input data
        seq_records = self._process_input_data(input_data)

        # Check for not-allowed characters in sequence ids
        if not store_by_hash:
            all_seq_ids_allowed = all(["/" not in seq_record.seq_id for seq_record in seq_records])
            if not all_seq_ids_allowed:
                raise ValueError("A sequence id contains the not allowed '/' character")

        # Sort sequences by length in descending order
        seq_records = list(sorted(seq_records,
                                  key=lambda seq_record: len(seq_record.seq),
                                  reverse=True))

        start_time = time.time()

        # Open the h5 file once and write embeddings as they're generated
        with h5py.File(embeddings_file_path, "a") as embeddings_file:
            for seq_record, embedding in tqdm(
                    self._embeddings_generator(seq_records, use_reduced_embeddings),
                    total=len(seq_records),
                    desc="Computing Embeddings",
                    disable=is_running_in_notebook()
            ):
                self.store_embedding(embeddings_file_handle=embeddings_file,
                                     seq_record=seq_record,
                                     embedding=embedding,
                                     store_by_hash=store_by_hash)

        end_time = time.time()
        logger.info(f"Time elapsed for computing embeddings: {end_time - start_time:.2f}[s]")

        return str(embeddings_file_path)

    @staticmethod
    def store_embedding(embeddings_file_handle, seq_record, embedding, store_by_hash: bool = True):
        h5_index = seq_record.get_hash() if store_by_hash else seq_record.seq_id
        embeddings_file_handle.create_dataset(h5_index, data=embedding.cpu().numpy(), compression="gzip", chunks=True)
        embeddings_file_handle[h5_index].attrs["original_id"] = seq_record.seq_id

    def generate_embeddings(self,
                            input_data: Union[str, Path, List[str], List[BiotrainerSequenceRecord], Dict[
                                str, BiotrainerSequenceRecord]],
                            reduce: bool) -> Generator[Tuple[BiotrainerSequenceRecord, torch.tensor], None, None]:
        """
        Generator function that yields embeddings as they are computed.

        Parameters:
            input_data: Input sequences in various formats
            reduce: If True, embeddings will be reduced to per-sequence embeddings.

        Yields:
            Tuple[BiotrainerSequenceRecord, torch.tensor]: Tuple of (BiotrainerSequenceRecord, embedding)
        """

        # Process input data
        seq_records = self._process_input_data(input_data)

        # Sort sequences by length in descending order
        seq_records = list(sorted(seq_records,
                                  key=lambda seq_record: len(seq_record.seq),
                                  reverse=True))

        # Generate embeddings
        yield from self._embeddings_generator(seq_records, reduce)

    @staticmethod
    def _process_input_data(input_data) -> List[BiotrainerSequenceRecord]:
        """
        Process various input formats into a list of BiotrainerSequenceRecord
        """
        if isinstance(input_data, (str, Path)):
            return [seq_record for seq_record in read_FASTA(input_data)]
        elif isinstance(input_data, list):
            if isinstance(input_data[0], BiotrainerSequenceRecord):
                return input_data
            elif isinstance(input_data[0], str):
                return [BiotrainerSequenceRecord(seq_id=f"Seq{idx}", seq=seq)
                        for idx, seq in enumerate(input_data)]
            else:
                raise ValueError(f"Non-supported type for input_data: {type(input_data[0])}")
        elif isinstance(input_data, dict):
            return [seq_record for seq_record in input_data.values()]
        else:
            raise ValueError(f"Non-supported type for input_data: {type(input_data)}")

    def _embeddings_generator(self,
                              seq_records: List[BiotrainerSequenceRecord],
                              use_reduced_embeddings: bool) \
            -> Generator[Tuple[BiotrainerSequenceRecord, torch.tensor], None, None]:
        """
        Core embedding computation logic that can be used by both save and generate methods

        Parameters:
            seq_records: List of sequence records to process
            use_reduced_embeddings: Whether to reduce embeddings to per-protein

        Yields:
            Tuple[BiotrainerSequenceRecord, torch.tensor]: Tuple of (BiotrainerSequenceRecord, embedding)
        """
        sequences = [seq_record.seq for seq_record in seq_records]
        embedding_iter = self._embedder.embed_many(sequences)

        for seq_record, embedding in zip(seq_records, embedding_iter):
            if embedding is None:
                raise Exception("Encountered None value during embedding calculation!")

            if use_reduced_embeddings:
                # TODO Batching might improve speed here
                embedding = self._embedder.reduce_per_protein(embedding)

            yield seq_record, embedding

    @staticmethod
    def get_embeddings_file_path(output_dir: Path,
                                 protocol: Protocol,
                                 embedder_name: str,
                                 use_half_precision: bool,
                                 force_output_dir: Optional[bool] = False) -> Path:
        use_reduced_embeddings = protocol in Protocol.using_per_sequence_embeddings()
        embedder_name = embedder_name.split("/")[-1]

        if force_output_dir:
            embeddings_file_path = output_dir
        else:
            # Create protocol path to embeddings
            embeddings_file_path = output_dir / protocol.name
            if not os.path.isdir(embeddings_file_path):
                os.mkdir(embeddings_file_path)

            # Add embedder name subdirectory
            embeddings_file_path /= embedder_name
            if not os.path.isdir(embeddings_file_path):
                os.mkdir(embeddings_file_path)

        # Append file name to output path
        embeddings_file_path /= EmbeddingService.get_embeddings_file_name(embedder_name,
                                                                          use_half_precision,
                                                                          use_reduced_embeddings)
        return embeddings_file_path

    @staticmethod
    def get_embeddings_file_name(embedder_name: str,
                                 use_half_precision: bool,
                                 use_reduced_embeddings: bool):
        embedder_name = embedder_name.split("/")[-1]
        return (("reduced_" if use_reduced_embeddings else "")
                + f"embeddings_file_{embedder_name}{'_half' if use_half_precision else ''}.h5")

    @staticmethod
    def embeddings_dimensionality_reduction(
            embeddings: Dict[str, torch.tensor],
            dimension_reduction_method: str,
            n_reduced_components: int) -> Dict[str, torch.tensor]:
        """Reduces the dimension of per-protein embeddings using one of the
        dimensionality reduction methods

        Args:
            embeddings (Dict[str, torch.tensor]): Dictionary of embeddings.
            dimension_reduction_method (str): The method used to reduce
            the dimensionality of embeddings. Options are 'umap' or 'tsne'.
            n_reduced_components (int): The target number of dimensions for
            the reduced embeddings.

        Returns:
            Dict[str, torch.tensor]: Dictionary of embeddings with reduced dimensions.
        """
        from umap import UMAP

        sorted_keys = sorted(list(embeddings.keys()))
        all_embeddings = torch.stack([embeddings[k] for k in sorted_keys], dim=0)
        max_dim_dict = {
            "umap": all_embeddings.shape[0] - 2,
            "tsne": all_embeddings.shape[0] - 1
        }
        n_reduced_components = min([
            n_reduced_components,
            max_dim_dict[dimension_reduction_method],
            all_embeddings.shape[1]])
        dimension_reduction_method_dict = {
            "umap": UMAP(n_components=n_reduced_components),
            "tsne": TSNE(
                n_components=n_reduced_components,
                perplexity=min(30, n_reduced_components))
        }
        logger.info(f"Starting embeddings dimensionality reduction via method {dimension_reduction_method}")
        embeddings_reduced_dimensions = dimension_reduction_method_dict[
            dimension_reduction_method].fit_transform(all_embeddings)
        logger.info(f"Finished embeddings dimensionality reduction!")
        return {sorted_keys[i]: torch.tensor(embeddings_reduced_dimensions[i]) for i in range(len(sorted_keys))}

    @staticmethod
    def load_embeddings(embeddings_file_path: str) -> Dict[str, torch.tensor]:
        """
        Loads precomputed embeddings from a file.

        Parameters:
            embeddings_file_path (str): Path to the embeddings file.

        Returns:
            out (Dict[str, torch.tensor]): Dictionary mapping sequence hashes to embeddings.
        """
        # Load computed embeddings in .h5 file format
        logger.info(f"Loading embeddings from: {embeddings_file_path}")
        start = time.perf_counter()

        # Old version see:
        # https://stackoverflow.com/questions/48385256/optimal-hdf5-dataset-chunk-shape-for-reading-rows/48405220#48405220
        embeddings_file = h5py.File(embeddings_file_path, 'r')

        # Sequence hash from embeddings file -> Embedding
        id2emb = {idx: torch.tensor(embedding).float() for (idx, embedding) in embeddings_file.items()}

        # Logging
        logger.info(f"Read {len(id2emb)} entries.")
        logger.info(f"Time elapsed for reading embeddings: {(time.perf_counter() - start):.1f}[s]")

        return id2emb



import time
import datetime

from ..pipeline import PipelineContext, PipelineStep
from ..pipeline.pipeline_step import PipelineStepType

from ...utilities import get_logger

logger = get_logger(__name__)


class PostProcessStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.POST_PROCESS

    @staticmethod
    def _onnx_export(context: PipelineContext):
        try:
            context.best_split.solver.save_as_onnx(embedding_dimension=context.n_features)
        except Exception:
            logger.error("Could not save model as ONNX!")

    @staticmethod
    def _save_finetuned_model(context: PipelineContext):
        embedding_service = context.embedding_service
        if embedding_service:
            embedding_service.save_embedder(output_dir=context.config["log_dir"])

    @staticmethod
    def _log_time(context: PipelineContext):
        pipeline_end_time = time.perf_counter()
        pipeline_end_time_abs = str(datetime.datetime.now().isoformat())
        pipeline_elapsed_time = pipeline_end_time - context.pipeline_start_time
        logger.info(f"Pipeline end time: {pipeline_end_time_abs}")
        logger.info(f"Total elapsed time for pipeline: {pipeline_elapsed_time} [s]")
        context.output_manager.add_derived_values({'pipeline_end_time': pipeline_end_time_abs})
        context.output_manager.add_derived_values({'pipeline_elapsed_time': pipeline_elapsed_time})

        logger.info(f"Extensive output information can be found at {context.config['output_dir']}/out.yml")

    def process(self, context: PipelineContext) -> PipelineContext:
        assert context.best_split is not None, f"Best split cannot be None at postprocess step!"

        # SAVE BEST SPLIT AS ONNX
        self._onnx_export(context)

        # SAVE FINETUNED EMBEDDER MODEL (IF APPLICABLE)
        self._save_finetuned_model(context)

        # LOG TIME
        self._log_time(context)

        return context



import torch

from scipy.stats import pearsonr
from typing import Dict, List, Any
from torch.utils.data import DataLoader

from .bootstrapper import Bootstrapper

from ..losses import get_loss
from ..models import get_model
from ..protocols import Protocol
from ..optimizers import get_optimizer
from ..solvers import MetricsCalculator, get_solver
from ..utilities import INTERACTION_INDICATOR, get_logger

logger = get_logger(__name__)


class SanityChecker:
    def __init__(self,
                 training_config: Dict[str, Any],
                 n_classes: int,
                 n_features: int,
                 metrics_calculator: MetricsCalculator,
                 train_dataset: List,
                 val_dataset: List,
                 test_dataset: List,
                 test_loader: DataLoader,
                 test_results_dict: Dict[str, Any],
                 mode: str = "warn"):
        self.training_config = training_config
        self.protocol = self.training_config.get("protocol")
        self.n_classes = n_classes
        self.n_features = n_features

        self.device = self.training_config.get("device")
        self.interaction = self.training_config.get("interaction", None)
        self.bootstrapping_iterations = self.training_config.get("bootstrapping_iterations", 0)
        if self.bootstrapping_iterations <= 0:
            self.bootstrapping_iterations = 30  # Always use bootstrapping for sanity checks with default 30 iterations
        self.metrics_calculator = metrics_calculator

        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.test_dataset = test_dataset
        self.test_loader = test_loader
        self.test_results_dict = test_results_dict

        self.mode = mode

        self._warnings = []

    def _handle_sanity_check_warning(self, warning: str):
        self._warnings.append(warning)
        if self.mode == "warn":
            logger.warning(warning)
        elif self.mode == "error":
            #  Might be useful for integration tests later
            raise SanityException(warning)

    def check_test_results(self, test_set_id) -> (Dict[str, Any], List[str]):
        logger.info(f"Running sanity checks on test set results ({test_set_id}) ..")

        self._check_metrics()
        self._check_predictions()
        baseline_dict = self._check_baselines()

        logger.info(f"Sanity check on test results ({test_set_id}) finished!")

        return baseline_dict, self._warnings

    def _check_metrics(self):
        if self.protocol in Protocol.classification_protocols():
            if "metrics" in self.test_results_dict.keys():
                test_result_metrics = self.test_results_dict['metrics']
            else:
                self._handle_sanity_check_warning(f"No test result metrics found!")
                return
            # Multi-class metrics
            if self.n_classes > 2:
                pass
            else:
                # Binary metrics
                accuracy = test_result_metrics['accuracy']
                precision = test_result_metrics['precision']
                recall = test_result_metrics['recall']
                if accuracy == precision == recall:
                    self._handle_sanity_check_warning(f"Accuracy ({accuracy}) == Precision == Recall for binary prediction!")

    def _check_predictions(self):
        mapped_predictions = self.test_results_dict.get('mapped_predictions', None)

        if mapped_predictions:
            predictions = list(mapped_predictions.values())
            # Check if the model is only predicting the same value for all test samples:
            if all(prediction == predictions[0] for prediction in predictions):
                self._handle_sanity_check_warning(f"Model is only predicting {predictions[0]} for all test samples!")

    def _check_baselines(self) -> Dict[str, Any]:
        baseline_dict = {"random_model": self._random_model_initialization_baseline()}

        if self.protocol in Protocol.classification_protocols():
            # Only for binary classification tasks at the moment:
            if self.n_classes <= 2:
                baseline_dict["one_only"] = self._one_only_baseline()
                baseline_dict["zero_only"] = self._zero_only_baseline()

                if self.interaction:
                    baseline_dict["bias_predictions"] = self._bias_interaction_baseline()

        # TODO residue_to_value
        elif self.protocol in Protocol.regression_protocols() and not self.protocol == Protocol.residue_to_value:
            baseline_dict["mean_only"] = self._mean_only_baseline()

        return baseline_dict

    def _one_only_baseline(self):
        """
        Predicts "1" for every sample in the test set. (Only for binary classification)
        """
        if self.protocol in Protocol.per_sequence_protocols() and self.protocol in Protocol.classification_protocols():
            one_only_baseline = self._value_only_baseline(value=1)
            logger.info(f"One-Only Baseline: {one_only_baseline}")
            return one_only_baseline

    def _zero_only_baseline(self):
        """
        Predicts "0" for every sample in the test set. (Only for binary classification)
        """
        if self.protocol in Protocol.per_sequence_protocols() and self.protocol in Protocol.classification_protocols():
            zero_only_baseline = self._value_only_baseline(value=0)
            logger.info(f"Zero-Only Baseline: {zero_only_baseline}")
            return zero_only_baseline

    def _mean_only_baseline(self):
        """
        Predicts the mean of the train set for every sample in the test set. (Only for regression)
        """
        if self.protocol in Protocol.regression_protocols():
            train_set_targets = torch.tensor([sample.target for sample in self.train_dataset])
            train_set_mean = torch.mean(train_set_targets).item()
            mean_only_baseline = self._value_only_baseline(value=train_set_mean)
            logger.info(f"Mean-Only Baseline: {mean_only_baseline}")
            return mean_only_baseline

    def _value_only_baseline(self, value: float):
        test_set_value_predictions = {sample.seq_id: value for sample in self.test_dataset}
        value_only_baseline = Bootstrapper.bootstrap(protocol=self.protocol,
                                                     device=self.device,
                                                     bootstrapping_iterations=self.bootstrapping_iterations,
                                                     metrics_calculator=self.metrics_calculator,
                                                     mapped_predictions=test_set_value_predictions,
                                                     test_loader=self.test_loader)
        return value_only_baseline

    def _random_model_initialization_baseline(self):
        model = get_model(n_classes=self.n_classes, n_features=self.n_features,
                          model_weights_init="normal", **self.training_config)
        optimizer = get_optimizer(protocol=self.protocol,
                                  model_parameters=model.parameters(),
                                  learning_rate=0.01,  # Model not trained, so parameter is not relevant
                                  optimizer_choice='adam',
                                  **{})
        loss = get_loss(**self.training_config)
        solver = get_solver(name="random_init_model", network=model,
                            optimizer=optimizer, loss_function=loss,
                            n_classes=self.n_classes, **self.training_config)
        random_init_inference = solver.inference(dataloader=self.test_loader, calculate_test_metrics=True)
        random_init_bootstrapping = Bootstrapper.bootstrap(protocol=self.protocol,
                                                           device=self.device,
                                                           bootstrapping_iterations=self.bootstrapping_iterations,
                                                           metrics_calculator=solver.metrics_calculator,
                                                           mapped_predictions=random_init_inference[
                                                               "mapped_predictions"],
                                                           test_loader=self.test_loader)
        logger.info(f"Random-Model Baseline: {random_init_bootstrapping}")
        return random_init_bootstrapping

    def _bias_interaction_baseline(self):
        """
            Calculates a dataset bias baseline for interactions (see for example
            Park, Marcotte 2011: https://doi.org/10.1093/bioinformatics/btr514).
            At first, it is counted how often a protein (id) is found in the positive and negative sets.
            Then, the dataset bias can be determined by calculating pearson-r for (positive_counts vs. negative counts).

            The predictor itself just sums up the positive and negative counts for both interactors and "predicts"
            the higher value.
        """
        if self.protocol in Protocol.per_sequence_protocols() and self.protocol in Protocol.classification_protocols():
            # 1. Calculate protein counts
            positive_counts = {}
            negative_counts = {}

            for sample in (self.train_dataset + self.val_dataset + self.test_dataset):
                interactor1 = sample.seq_id.split(INTERACTION_INDICATOR)[0]
                interactor2 = sample.seq_id.split(INTERACTION_INDICATOR)[1]
                for count_dict in [positive_counts, negative_counts]:
                    if interactor1 not in count_dict:
                        count_dict[interactor1] = 0
                    if interactor2 not in count_dict:
                        count_dict[interactor2] = 0

                if sample.target == 1:
                    positive_counts[interactor1] += 1
                    positive_counts[interactor2] += 1
                else:
                    negative_counts[interactor1] += 1
                    negative_counts[interactor2] += 1

            # 2. Calculate dataset bias
            dataset_bias = pearsonr(list(positive_counts.values()), list(negative_counts.values()))

            def bias_predictor(int1, int2):
                pos_occurrences_total = positive_counts[int1]
                pos_occurrences_total += positive_counts[int2]
                neg_occurrences_total = negative_counts[int1]
                neg_occurrences_total += negative_counts[int2]

                return 0 if neg_occurrences_total >= pos_occurrences_total else 1

            # 3. Predict all test_samples from bias
            predictions = []
            test_set_targets = []
            for test_sample in self.test_dataset:
                interactor1 = test_sample.seq_id.split(INTERACTION_INDICATOR)[0]
                interactor2 = test_sample.seq_id.split(INTERACTION_INDICATOR)[1]
                predictions.append(bias_predictor(interactor1, interactor2))
                test_set_targets.append(test_sample.target)

            # 4. Calculate metrics for bias predictions
            bias_metrics = self.metrics_calculator.compute_metrics(predicted=torch.tensor(predictions),
                                                                   labels=torch.tensor(test_set_targets))
            logger.info(f"Bias Baseline for interactions: {bias_metrics}")
            logger.info(f"Dataset bias for interactions: {dataset_bias}")
            return {"dataset_bias": {"bias": float(dataset_bias.statistic), "pvalue": float(dataset_bias.pvalue)},
                    **bias_metrics}


class SanityException(Exception):
    pass



import torch
import psutil

from typing import Union


def get_device(device: Union[None, str, torch.device] = None) -> torch.device:
    """Returns what the user specified, or defaults to the GPU,
    with a fallback to CPU if no GPU is available."""
    if isinstance(device, torch.device):
        if device.type == "cuda" and torch.cuda.is_available():
            return device
        elif device.type == "mps" and torch.backends.mps.is_available():
            return device
        else:
            return torch.device("cpu")
    elif isinstance(device, str):
        if "cuda" in device and torch.cuda.is_available():
            return torch.device(device)
        elif "mps" in device and torch.backends.mps.is_available():
            return torch.device(device)
        else:
            return torch.device("cpu")
    elif torch.cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    else:
        return torch.device("cpu")


def is_device_cpu(device: Union[None, str, torch.device] = None) -> bool:
    if device is None:
        return False
    if isinstance(device, torch.device):
        return device.type == "cpu"
    if device is str:
        return device == "cpu"
    return False


def is_device_cuda(device: Union[None, str, torch.device] = None) -> bool:
    if device is None:
        return False
    if not torch.cuda.is_available():
        return False
    if isinstance(device, torch.device):
        return device.type == "cuda"
    if device is str:
        return "cuda" in device
    return False


def is_device_mps(device: Union[None, str, torch.device] = None) -> bool:
    if device is None:
        return False
    if not torch.backends.mps.is_available():
        return False
    if isinstance(device, torch.device):
        return device.type == "mps"
    if device is str:
        return "mps" in device
    return False


def get_device_memory(device: Union[None, str, torch.device] = None) -> float:
    """ Returns the amount of memory available for this device in GB. If it was not possible to calculate, 4 GB is
     used as a conservative default and 8 GB for mps """
    gb_factor = (1024 ** 3)
    conservative_default = 4
    default_mps = 8  # TODO [Cross platform] Improve MacOS RAM estimation

    if device is None or isinstance(device, str):
        device = get_device()
    if is_device_cuda(device):
        free_mem, _ = torch.cuda.mem_get_info(device=device)
        return free_mem / gb_factor
    if is_device_cpu(device):
        vm = psutil.virtual_memory()
        return vm.available / gb_factor
    if is_device_mps(device):
        return default_mps
    return conservative_default



import unittest

from pathlib import Path
from copy import deepcopy
from biotrainer.config import Configurator, ConfigurationException

configurations = {
    "minimal": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
    },
    "required": {
        # Missing input file
        "protocol": "residue_to_class"
    },
    "download": {
        "input_file": "https://example.com/sequences.fasta",
        "protocol": "sequence_to_class",
        "model_choice": "FNN",
    },
    "download_prohibited": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
        "model_choice": "FNN",
        "embedder_name": "https://example.com/payload.py"
    },
    "auto_resume_pretrained_model_mutual_exclusive": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
        "auto_resume": True,
        "pretrained_model": "placeholder.pt"
    },
    "mutual_exclusive_device_half_precision": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
        "device": "cpu",
        "use_half_precision": True
    },
    "local_custom_embedder": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
        "embedder_name": "../examples/custom_embedder/ankh/ankh_embedder.py"
    },
    "file_paths": {
        "input_file": "test_input_files/r2c/input.fasta",
        "embedder_name": "../examples/custom_embedder/esm2/esm2_embedder.py",
        "output_dir": "test_output",
        "protocol": "residue_to_class"
    },
    "multiple_values": {
        "input_file": "test_input_files/r2c/input.fasta",
        "protocol": "sequence_to_class",
        "model_choice": ["FNN", "DeeperFNN"],
        "optimizer_choice": ["adam", "sgd"],
        "num_epochs": [100, 200],
        "learning_rate": [1e-3, 1e-4],
        "batch_size": [64, 128],
        "epsilon": [1e-4, 1e-3],
        "patience": [10, 20],
        "use_class_weights": [True, False]
    },
    "hold_out": {
        "cross_validation_config": {
            "method": "hold_out",
            "choose_by": "loss"
        }
    },
    "k_fold": {
        "cross_validation_config": {
            "method": "k_fold",
            "k": 3,
            "stratified": True,
            "repeat": 1
        }
    },
    "nested_k_fold": {
        "cross_validation_config": {
            "method": "k_fold",
            "repeat": 2,
            "k": 4,
            "nested": True,
            "nested_k": 3,
            "search_method": "random_search",
            "n_max_evaluations_random": 5
        }
    },
    "leave_p_out": {
        "cross_validation_config": {
            "method": "leave_p_out",
            "p": 5,
        }
    },
    "hf_valid_for_sequences": {
        "protocol": "sequence_to_class",
        "hf_dataset": {
            "path": "heispv/protein_data_test",
            "subset": "split_1",
            "sequence_column": "protein_sequence",
            "target_column": "protein_class"
        }
    },
    "hf_valid_for_residues": {
        "protocol": "residue_to_class",
        "hf_dataset": {
            "path": "heispv/protein_data_test",
            "subset": "split_3",
            "sequence_column": "protein_sequence",
            "target_column": "secondary_structure"
        }
    },
    "hf_no_subset_required": {
        "protocol": "residue_to_class",
        "hf_dataset": {
            "path": "heispv/protein_data_test_2",
            "subset": "random_subset_name",
            "sequence_column": "protein_sequence",
            "target_column": "secondary_structure"
        }
    },
    "finetuning_ohe": {
        "protocol": "sequence_to_class",
        "embedder_name": "one_hot_encoding",
        "finetuning_config": {
            "method": "lora",
        }
    },
    "finetuning_no_method": {
        "protocol": "sequence_to_class",
        "embedder_name": "one_hot_encoding",
        "finetuning_config": {
            "lora_r": 10,
        }
    }
}


class ConfigurationVerificationTests(unittest.TestCase):

    def test_minimal_configuration(self):
        configurator = Configurator.from_config_dict(configurations["minimal"])
        self.assertTrue(configurator.get_verified_config(), "Minimal config does not work!")

    def test_required(self):
        configurator = Configurator.from_config_dict(configurations["required"])
        with self.assertRaises(ConfigurationException,
                               msg="Config with missing required config option does not throw an error!"):
            configurator.get_verified_config()

    def test_wrong_model(self):
        config_dict = configurations["minimal"]
        config_dict["model_choice"] = "RNN123"
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException,
                               msg="Config with arbitrary model does not throw an exception"):
            configurator.get_verified_config()

        config_dict["model_choice"] = "LightAttention"
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException,
                               msg="Config with unavailable model for the protocol does not throw an exception"):
            configurator.get_verified_config()

    def test_download(self):
        config_dict = configurations["download"]
        configurator = Configurator.from_config_dict(config_dict)

        # Download allowed but must not work from arbitrary URL
        with self.assertRaisesRegex(Exception, expected_regex="Could not download",
                                    msg="Config downloading a sequence file does not throw Exception"):
            configurator.get_verified_config()

    def test_download_prohibited(self):
        config_dict = configurations["download_prohibited"]
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException,
                               msg="Config downloading an embedding script does not throw an exception"):
            configurator.get_verified_config()

    def test_auto_resume_pretrained_model_mutual_exclusive(self):
        config_dict = configurations["auto_resume_pretrained_model_mutual_exclusive"]
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaisesRegex(ConfigurationException, expected_regex="mutual exclusive",
                                    msg="Config with auto_resume and pretrained_model does not throw an error"):
            configurator.get_verified_config()

    def test_mutual_exclusive_cpu_half_precision(self):
        configurator = Configurator.from_config_dict(configurations["mutual_exclusive_device_half_precision"])
        with self.assertRaises(ConfigurationException,
                               msg="Config with prohibited config combination does not throw an error!"):
            configurator.get_verified_config()

    def test_local_custom_embedder(self):
        configurator = Configurator.from_config_dict(configurations["local_custom_embedder"])
        self.assertTrue(configurator.get_verified_config(), "Local custom embedder config does not work!")

    def test_file_paths_absolute_after_verification(self):
        config_file_paths = configurations["file_paths"]
        configurator = Configurator.from_config_dict(config_file_paths)
        verified_config = configurator.get_verified_config()
        for key in config_file_paths.keys():
            if key != "protocol" and "file" in key:
                self.assertTrue(Path(verified_config[key]).is_absolute(),
                                "File paths not absolute after verification!")

    def test_hold_out(self):
        config_dict = {**configurations["minimal"], **configurations["hold_out"]}
        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "Hold out cross validation does not work!")

    def test_k_fold(self):
        config_dict = {**configurations["minimal"], **configurations["k_fold"]}
        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "k_fold does not work!")

        with self.assertRaises(ConfigurationException,
                               msg="Config with missing k_fold params does not throw an exception"):
            config_dict["cross_validation_config"].pop("k")
            configurator.from_config_dict(config_dict).get_verified_config()

    def test_nested_k_fold(self):
        config_dict = {**configurations["multiple_values"], **configurations["nested_k_fold"]}
        original_config = deepcopy(config_dict)

        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "nested k_fold does not work!")

        with self.assertRaises(ConfigurationException,
                               msg="Config with missing nested k_fold param does not throw an exception"):
            config_dict["cross_validation_config"].pop("nested_k")
            configurator.from_config_dict(config_dict).get_verified_config()

        config_dict = deepcopy(original_config)
        with self.assertRaises(ConfigurationException,
                               msg="Config with missing search method for nested k_fold does not throw an exception"):
            config_dict["cross_validation_config"].pop("search_method")
            configurator.from_config_dict(config_dict).get_verified_config()

    def test_leave_p_out(self):
        config_dict = {**configurations["minimal"], **configurations["leave_p_out"]}
        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "leave_p_out does not work!")

        with self.assertRaises(ConfigurationException,
                               msg="Config with missing p param does not throw an exception"):
            config_dict["cross_validation_config"].pop("p")
            configurator.from_config_dict(config_dict).get_verified_config()

    def test_multiple_values(self):
        config_dict = {**configurations["multiple_values"], **configurations["nested_k_fold"],
                       "learning_rate": "[10**-x for x in [2, 3, 4]]"}
        # Test list comprehension
        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "List comprehension for hyperparameter optimization"
                                                            "does not work!")
        # Test range expression
        config_dict["batch_size"] = "range(64, 132, 4)"
        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(configurator.get_verified_config(), "Range expression for hyperparameter optimization"
                                                            "does not work!")
        # Check that with missing nested k fold cross validation, optimization does not work
        config_dict = {**configurations["multiple_values"], **configurations["hold_out"]}
        with self.assertRaises(ConfigurationException,
                               msg="Config with multiple values for hold_out cv does not "
                                   "throw an exception!"):
            configurator.from_config_dict(config_dict).get_verified_config()

        config_dict = {**configurations["multiple_values"], **configurations["k_fold"]}
        with self.assertRaises(ConfigurationException,
                               msg="Config with multiple values for k_fold (not nested) cv does not "
                                   "throw an exception!"):
            configurator.from_config_dict(config_dict).get_verified_config()

        config_dict = {**configurations["multiple_values"], **configurations["leave_p_out"]}
        with self.assertRaises(ConfigurationException,
                               msg="Config with multiple values for leave_p_out cv does not throw an exception!"):
            configurator.from_config_dict(config_dict).get_verified_config()

    def test_hf_valid_3_split(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])

        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(
            configurator.get_verified_config(),
            "Valid hf_dataset configuration for sequences failed."
        )

    def test_hf_invalid_1_split(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["hf_dataset"]["subset"] = "split_2"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_valid_residues_protocol(self):
        config_dict = deepcopy(configurations["hf_valid_for_residues"])

        configurator = Configurator.from_config_dict(config_dict)
        self.assertTrue(
            configurator.get_verified_config(),
            "Valid hf_dataset configuration for residues failed."
        )

    def test_hf_missing_sequence_column_name(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        del config_dict["hf_dataset"]["sequence_column"]

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_missing_labels_column_name(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        del config_dict["hf_dataset"]["target_column"]

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_invalid_sequence_column_name(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["hf_dataset"]["sequence_column"] = "random_invalid_name"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(Exception) as context:
            configurator.get_verified_config()

    def test_hf_invalid_target_column_name(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["hf_dataset"]["target_column"] = "random_invalid_name"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_invalid_path(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["hf_dataset"]["path"] = "random_invalid_name"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_invalid_subset(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["hf_dataset"]["subset"] = "random_invalid_name"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_requires_subset(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        del config_dict["hf_dataset"]["subset"]

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_hf_not_requires_subset(self):
        config_dict = deepcopy(configurations["hf_no_subset_required"])

        configurator = Configurator.from_config_dict(config_dict)
        print(configurator._config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()


    def test_hf_mutual_exclusive_input_file_name(self):
        config_dict = deepcopy(configurations["hf_valid_for_sequences"])
        config_dict["input_file"] = "random_invalid_name"

        configurator = Configurator.from_config_dict(config_dict)
        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()


    def test_dimension_reduction_methods(self):
        config_dict = deepcopy(configurations["minimal"])
        # Existing method works
        config_dict["dimension_reduction_method"] = "umap"

        configurator = Configurator.from_config_dict(config_dict)

        self.assertTrue(
            configurator.get_verified_config(),
            "Valid dimension_reduction_method: umap failed!"
        )

        # Non-existing method does not work
        config_dict["dimension_reduction_method"] = "nonexistingmethod"
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_dimension_reduction_components(self):
        config_dict = deepcopy(configurations["minimal"])
        config_dict["dimension_reduction_method"] = "umap"
        # Positive integers are valid
        valid_values = [22, 1, 44, 123, 99, 1000]
        for valid_value in valid_values:
            config_dict["n_reduced_components"] = valid_value

            configurator = Configurator.from_config_dict(config_dict)
            self.assertTrue(
                configurator.get_verified_config(),
                f"Valid n_reduced_components: {valid_value} failed"
            )

        # Other values are not valid
        invalid_values = [0, -50, 5.5, -10.25, 33.999, 44.000]
        for invalid_value in invalid_values:
            config_dict["n_reduced_components"] = invalid_value
            configurator = Configurator.from_config_dict(config_dict)

            with self.assertRaises(ConfigurationException) as context:
                configurator.get_verified_config()

    def test_bootstrapping_iterations(self):
        config_dict = deepcopy(configurations["minimal"])
        # Positive integers and zero are valid
        valid_values = [0, 1, 44, 123]
        for valid_value in valid_values:
            config_dict["bootstrapping_iterations"] = valid_value

            configurator = Configurator.from_config_dict(config_dict)
            self.assertTrue(
                configurator.get_verified_config(),
                f"Valid bootstrapping_iterations: {valid_value} failed"
            )

        # Other values are not valid
        invalid_values = [-1, 5.5, -2.1, 0.1, 1.00000]
        for invalid_value in invalid_values:
            config_dict["bootstrapping_iterations"] = invalid_value
            configurator = Configurator.from_config_dict(config_dict)

            with self.assertRaises(ConfigurationException) as context:
                configurator.get_verified_config()

    def test_finetuning_ohe(self):
        config_dict = deepcopy(configurations["finetuning_ohe"])
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()

    def test_finetuning_no_method(self):
        config_dict = deepcopy(configurations["finetuning_no_method"])
        configurator = Configurator.from_config_dict(config_dict)

        with self.assertRaises(ConfigurationException) as context:
            configurator.get_verified_config()



from .setup_step import SetupStep
from .testing_step import TestingStep
from .data_loading_step import DataLoadingStep
from .postprocess_step import PostProcessStep
from .projection_step import ProjectionStep
from .training_step import TrainingStep
from .input_validation_step import InputValidationStep
from .embedding_step import EmbeddingStep, FineTuningEmbeddingStep

__all__ = [
    "SetupStep",
    "InputValidationStep",
    "TestingStep",
    "DataLoadingStep",
    "EmbeddingStep",
    "FineTuningEmbeddingStep",
    "PostProcessStep",
    "ProjectionStep",
    "TrainingStep",
]


import shutil

from pathlib import Path
from typing import Any
from urllib import request
from urllib.parse import urlparse

from .config_exception import ConfigurationException


def download_file_from_config(option_name: str, url: str, config_file_path: Path):
    """
    Download the file from a URL if downloading is allowed.

    Args:
        option_name (str): The name of the option to download.
        url (str): The url to download the file from.
        config_file_path (Path): Path to the configuration file directory.
    Raises:
        Exception: If the file cannot be downloaded.
    """
    if is_url(url):
        try:
            print(f"Trying to download {option_name} from {url}")
            req = request.Request(url, headers={
                'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64)"
            })

            file_name = url.split("/")[-1]
            save_path = str(config_file_path / f"downloaded_{file_name}")
            with request.urlopen(req) as response, open(save_path, 'wb') as outfile:
                if response.status == 200:
                    print(f"OK - Downloading file {option_name} (size: {response.length / (2 ** 20):.2f} MB)..")
                shutil.copyfileobj(response, outfile)
            print(f"{option_name} successfully downloaded and stored at {save_path}.")
            return save_path
        except Exception as e:
            raise Exception(f"Could not download {option_name} from url {url}") from e
    else:
        raise ConfigurationException(f"Provided url {url} for {option_name} is not a valid url!")

def make_path_absolute_if_necessary(value: str, config_file_path: Path) -> str:
    """
    Convert the file path to an absolute path if it exists.

    Args:
        value (str): The file path.
        config_file_path (Path): Path to the configuration file directory.
    """
    absolute_path = (config_file_path / value).absolute()
    if absolute_path.is_file() or absolute_path.is_dir():
        return str(absolute_path)

    # File path is already absolute and not directly within the config directory
    absolute_value_path = Path(value).absolute()
    if absolute_value_path.is_file() or absolute_value_path.is_dir():
        return str(absolute_value_path)
    raise ConfigurationException(f"Could not find path for file {value}.")


def is_url(value: str) -> bool:
    """
    Check if the given value is a valid URL.

    Args:
        value (str): The value to check.

    Returns:
        bool: True if `value` is a URL with scheme http, https, or ftp, else False.
    """
    return urlparse(value).scheme in ["http", "https", "ftp"]


def is_list_option(value: Any) -> bool:
    """
    Determine if the option value is a list, range, or a valid range string.

    Used to check for hyperparameter optimization.

    Returns:
        bool: True if `value` is a list, range, or a valid range string, else False.
    """
    return ("range" in str(value) or isinstance(value, list) or
            (isinstance(value, str) and "[" in value and "]" in value))


import json

from pathlib import Path
from typing import Dict, Any, Union

from .data_handler import AutoEvalTask


class ReportManager:
    def __init__(self, embedder_name: str, training_date, min_seq_len: int, max_seq_len: int):
        self.embedder_name = embedder_name
        self.training_date = training_date
        self.min_seq_len = min_seq_len
        self.max_seq_len = max_seq_len
        self.results = {}

    def add_result(self, task: AutoEvalTask, result_dict: Dict[str, Any]):
        self.results[task.name] = result_dict

    def write(self, output_dir: Union[Path, str]) -> Dict[str, Any]:
        result_dict = {
            "embedder_name": self.embedder_name,
            "training_date": self.training_date,
            "min_seq_len": self.min_seq_len,
            "max_seq_len": self.max_seq_len,
            "results": self.results
        }
        report_name = output_dir / f'autoeval_report_{self.embedder_name}.json'

        print(f'Writing autoeval report to: {report_name}')
        with open(report_name, 'w') as report_file:
            report_file.write(json.dumps(result_dict))

        return result_dict



from __future__ import annotations

import re
import json

from pathlib import Path
from typing import Dict, Iterable, List

from transformers import PreTrainedTokenizer


class CustomTokenizer(PreTrainedTokenizer):
    def __init__(self, vocab: Dict[str, int],
                 unk_token: str = "<unk>", pad_token: str = "<pad>", eos_token: str = "</s>",
                 characters_to_replace: str = "UZOB",
                 replacement_character: str = "X",
                 uses_whitespaces: bool = False):
        # Define vocabulary mapping amino acids & special tokens
        self.vocab = vocab

        # Reverse vocabulary for decoding
        self.reverse_vocab = {v: k for k, v in self.vocab.items()}

        # Set special tokens explicitly
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.eos_token = eos_token

        # Set preprocessing
        self.preprocessing_strategy = lambda sequences: self._get_preprocessing_strategy(sequences,
                                                                                         characters_to_replace,
                                                                                         replacement_character,
                                                                                         uses_whitespaces)
        # Initialize parent class properly
        super().__init__(
            unk_token=unk_token,
            pad_token=pad_token,
            eos_token=eos_token,
            vocab=vocab
        )

    @staticmethod
    def _get_preprocessing_strategy(sequences: Iterable[str],
                                    characters_to_replace,
                                    replacement_character,
                                    uses_whitespaces) -> List[str]:
        # Remove replacement characters
        sequences_cleaned = [re.sub(fr"[{characters_to_replace}]", replacement_character, sequence)
                             for sequence in sequences]

        join_char = " " if uses_whitespaces else ""
        sequences_with_spaces = [join_char.join(list(sequence)) for sequence in sequences_cleaned]
        return sequences_with_spaces

    @classmethod
    def from_config(cls, config_path: Path) -> CustomTokenizer:
        """
        Initialize a CustomTokenizer from a config file.

        Args:
            config_path (Path): Path to the JSON config file

        Returns:
            CustomTokenizer: Initialized tokenizer instance
        """

        # Read the config file
        with open(config_path, 'r') as f:
            config = json.load(f)

        # Ensure vocab is present
        if 'vocab' not in config:
            raise ValueError("Tokenizer config file must contain 'vocab' dictionary")

        # Get required vocab
        vocab = config['vocab']
        if not isinstance(vocab, dict):
            raise ValueError("Tokenizer config file must contain 'vocab' dictionary")
        if not all(isinstance(key, str) for key in vocab.keys()):
            raise ValueError("'vocab' dictionary keys in tokenizer config must all be strings")
        if not all(isinstance(value, int) for value in vocab.values()):
            raise ValueError("'vocab' dictionary values in tokenizer config must all be integers")

        # Get optional parameters with defaults
        unk_token = config.get('unk_token', "<unk>")
        pad_token = config.get('pad_token', "<pad>")
        eos_token = config.get('eos_token', "</s>")
        characters_to_replace = config.get('characters_to_replace', "UZOB")
        replacement_character = config.get('replacement_character', "X")
        uses_whitespaces = config.get('uses_whitespaces', False)

        # Initialize and return the tokenizer
        return cls(
            vocab=vocab,
            unk_token=unk_token,
            pad_token=pad_token,
            eos_token=eos_token,
            characters_to_replace=characters_to_replace,
            replacement_character=replacement_character,
            uses_whitespaces=uses_whitespaces
        )

    def get_vocab(self):
        """ Returns the vocabulary dictionary. """
        return self.vocab

    @property
    def vocab_size(self):
        return len(self.vocab)

    def _tokenize(self, text, **kwargs):
        return list(text)

    def _convert_token_to_id(self, token):
        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index):
        # Reverse the vocab to convert id back to token
        reverse_vocab = {v: k for k, v in self.vocab.items()}
        return reverse_vocab.get(index, self.unk_token)



from ...protocols import Protocol

# Unused dataset splits are commented out

FLIP_DATASETS = {
    "aav": {
        "splits": [
            # "des_mut",
            # "mut_des",
            # "one_vs_many",
            "two_vs_many",
            # "seven_vs_many",
            "low_vs_high",
            # "sampled"
        ],
        "evaluation_metric": "spearmans-corr-coeff",
        "protocol": Protocol.sequence_to_value,
    },
    "bind": {
        "splits": [
            # "one_vs_many",
            # "two_vs_many",
            "from_publication",
            # "one_vs_sm",
            # "one_vs_mn",
            # "one_vs_sn"
        ],
        "evaluation_metric": "macro-f1_score",
        "protocol": Protocol.residue_to_class,
    },
    # "conservation": {
    #     "splits": [
    #         "sampled"
    #     ],
    #     "evaluation_metric": "accuracy",
    #     "protocol": Protocol.residue_to_class,
    # },
    "gb1": {
        "splits": [
            # "one_vs_rest",
            "two_vs_rest",
            # "three_vs_rest",
            "low_vs_high",
            # "sampled"
        ],
        "evaluation_metric": "spearmans-corr-coeff",
        "protocol": Protocol.sequence_to_value,
    },
    "meltome": {
        "splits": [
            "mixed_split",
            # "human",
            # "human_cell"
        ],
        "evaluation_metric": "spearmans-corr-coeff",
        "protocol": Protocol.sequence_to_value,
    },
    "scl": {
        "splits": [
            # "mixed_soft",
            "mixed_hard",
            # "human_soft",
            # "human_hard",
            # "balanced",
            # "mixed_vs_human_2"
        ],
        "evaluation_metric": "accuracy",
        "protocol": Protocol.sequence_to_class,
    },
    # "sav": {
    #     "splits": [
    #         "mixed",
    #         # "human",
    #         # "only_savs"
    #     ],
    #     "evaluation_metric": "f1_score",
    #     "protocol": Protocol.sequence_to_class,
    # },
    "secondary_structure": {
        "splits": [
            "sampled"
        ],
        "evaluation_metric": "accuracy",
        "protocol": Protocol.residue_to_class,
    },
}



import torch

from pathlib import Path
from typing import Union, Optional, List, Tuple, Dict, Any
from transformers import AutoTokenizer, T5Tokenizer, T5EncoderModel, EsmTokenizer, EsmModel, BertTokenizer, \
    BertForMaskedLM

from .onnx import OnnxEmbedder
from .huggingface import HuggingfaceTransformerEmbedder
from .interfaces import EmbedderInterface, CustomTokenizer
from .services import EmbeddingService, PeftEmbeddingService
from .predefined_embedders import RandomEmbedder, AAOntologyEmbedder, OneHotEncodingEmbedder, Blosum62Embedder

from ..utilities import is_device_cpu, get_logger

__PREDEFINED_EMBEDDERS = {
    "one_hot_encoding": OneHotEncodingEmbedder,
    "random_embedder": RandomEmbedder,
    "AAOntology": AAOntologyEmbedder,
    "blosum62": Blosum62Embedder,
}

logger = get_logger(__name__)


def get_embedding_service(embedder_name: str,
                          custom_tokenizer_config: Optional[str],
                          use_half_precision: Optional[bool] = False,
                          device: Optional[Union[str, torch.device]] = None,
                          finetuning_config: Optional[Dict[str, Any]] = None) -> Union[
    EmbeddingService, PeftEmbeddingService]:
    embedder: EmbedderInterface = _get_embedder(embedder_name=embedder_name,
                                                custom_tokenizer_config=custom_tokenizer_config,
                                                use_half_precision=use_half_precision,
                                                device=device)
    if finetuning_config:
        return PeftEmbeddingService(embedder=embedder, use_half_precision=use_half_precision,
                                    finetuning_config=finetuning_config)

    return EmbeddingService(embedder=embedder, use_half_precision=use_half_precision)


def _determine_tokenizer_and_model(embedder_name: str) -> Tuple:
    """
    Method to find the model and tokenizer architecture from the embedder name for huggingface transformers.
    AutoTokenizer does not work for all models, such that the next best option is to use the embedder name.

    @param embedder_name: Name of huggingface transformer
    @return: Tuple of tokenizer and model class
    """
    try:
        auto_tok = AutoTokenizer.from_pretrained(embedder_name)
        embedder_class = auto_tok.__class__.__name__
    except Exception:
        embedder_class = embedder_name

    if "t5" in embedder_class.lower():
        return T5Tokenizer, T5EncoderModel
    elif "esm" in embedder_class.lower():
        return EsmTokenizer, EsmModel
    elif "bert" in embedder_class.lower():
        return BertTokenizer, BertForMaskedLM

    # Use T5 as default
    return T5Tokenizer, T5EncoderModel


def _get_embedder(embedder_name: Optional[str],
                  custom_tokenizer_config: Optional[str],
                  use_half_precision: Optional[bool],
                  device: Optional[Union[str, torch.device]]) -> EmbedderInterface:
    if embedder_name is None or embedder_name == "":
        raise ValueError("No embedder name provided!")

    # Predefined Embedders
    if embedder_name in __PREDEFINED_EMBEDDERS.keys():
        logger.info(f"Using predefined embedder: {embedder_name}")
        return __PREDEFINED_EMBEDDERS[embedder_name]()

    # Custom Embedders
    if embedder_name.endswith(".py"):  # Check that file ends on .py => Python script
        raise ValueError(f"Custom embedders from script have been replaced with onnx models in v0.9.7!")

    # ONNX Embedders
    if embedder_name.endswith(".onnx"):
        if custom_tokenizer_config:
            tokenizer = CustomTokenizer.from_config(config_path=Path(custom_tokenizer_config))
            logger.info(f"Using custom tokenizer (vocab size: {tokenizer.vocab_size})!")
        else:
            tokenizer = T5Tokenizer.from_pretrained(embedder_name, dtype=torch.float32)
        onnx_model = OnnxEmbedder(onnx_path=Path(embedder_name), tokenizer=tokenizer, device=device)
        logger.info(f"Using onnx embedder: {onnx_model.name}")
        return onnx_model

    # Huggingface Transformer Embedders
    if use_half_precision and is_device_cpu(device):
        raise ValueError(f"use_half_precision mode is not compatible with embedding "
                         f"on the CPU. (See: https://github.com/huggingface/transformers/issues/11546)")

    dtype = torch.float16 if use_half_precision else torch.float32

    tokenizer_class, model_class = _determine_tokenizer_and_model(embedder_name)
    logger.info(f"Loading embedder model {embedder_name}..")
    try:
        tokenizer = tokenizer_class.from_pretrained(embedder_name, dtype=dtype)
        model = model_class.from_pretrained(embedder_name, dtype=dtype)
    except OSError as os_error:
        raise Exception(f"{embedder_name} could not be found!") from os_error
    except Exception:
        try:
            tokenizer = AutoTokenizer.from_pretrained(embedder_name, dtype=dtype)
            model = T5EncoderModel.from_pretrained(embedder_name, dtype=dtype)
        except Exception as e:
            raise Exception(f"Loading {embedder_name} automatically and as {tokenizer_class.__class__.__name__} failed!"
                            f" Please provide a custom_embedder script for your use-case.") from e

    logger.info(f"Using huggingface transformer embedder: {embedder_name} "
                f"- Model: {model.__class__.__name__} "
                f"- Tokenizer: {tokenizer.__class__.__name__} "
                f"- Half-Precision: {str(use_half_precision)}")
    model.to(device)
    return HuggingfaceTransformerEmbedder(name=embedder_name, model=model, tokenizer=tokenizer,
                                          use_half_precision=use_half_precision, device=device)


def get_predefined_embedder_names() -> List[str]:
    return list(__PREDEFINED_EMBEDDERS.keys())


__all__ = [
    "EmbeddingService",
    "PeftEmbeddingService",
    "OneHotEncodingEmbedder",
    "RandomEmbedder",
    "AAOntologyEmbedder",
    "get_embedding_service",
    "get_predefined_embedder_names"
]

